; Scripts for Skype 8 for Windows Desktop (Electron).
;
; Author: Doug Lee
/*
Copyright (c) 2017-2018 Doug Lee.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* The names of the copyright holders and contributors may not be used to
  endorse or promote products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

include "hjconst.jsh"
include "uia.jsh"
include "MSAAConst.jsh"
include "skypeelectron.jsh"
include "skypeelectron.jsm"

const scriptVer = 45

script scriptFileName()
scriptAndAppNames(formatString(msgScriptName, scriptVer))
endScript

void function autoStartEvent()
apphelp__setDocName("skypeelectron.htm")
mkcs__setCommandKey("[")
mkcs__setHelpText(msgMKCSHelpText)
endFunction

globals int skype8__giInCustomCode
script sayWindowTitle()
sutil__resetCache()
skype8__giInCustomCode = True
performScript sayWindowTitle()
skype8__giInCustomCode = False
sayMessage(OT_Message, sutil__getConvName())
endScript

int function IsBrowserContentWindow(handle hwnd)
; Makes sayWindowTitle report the root window name rather than the root MSAA/UIA document name.
; The window name includes an unread message count; the document name is just "Skype Preview." [DGL, 2018-05-03, Skype 8.20]
; This function is called during script sayWindowTitle by chrome.jss in at least JAWS 2018 update 26.
if skype8__giInCustomCode
	return False
endIf
return IsBrowserContentWindow(hwnd)
endFunction

void function sayObjectTypeAndText(optional int nLevel)
; Helps avoid some extra speech.
if getObjectValue(False, nLevel) == "file:///Index.html"
	; getObjectSubtypeCode should return WT_Document here but does not always do so.
	return
endIf
sayObjectTypeAndText(nLevel)
endFunction

script gotoHomePage()
sutil__gotoSkypeScriptPage("", False)
endScript

script sayAppVersion()
skype8__giInCustomCode = True
performScript sayAppVersion()
skype8__giInCustomCode = False
endScript

string function getVersionInfoString(string appPath, string sKey)
; Called by default sayAppVersion code.
var string result = getVersionInfoString(appPath, sKey)
if !skype8__giInCustomCode
|| sKey != "ProductVersion"
; The length restriction makes versions like "8.21" get ignored so we can get a more specific answer.
; Skype 8.21 (at least) returns versions like "8.21" for ProductVersion (used by default JAWS script)
; and the full version for FileVersion.  [DGL, 2018-05-22]
|| stringLength(result) >= 6
	return result
endIf
return getVersionInfoString(appPath, "FileVersion")
endFunction

int function supportsEditCallbacks()
; Part of avoiding "blank" on up/down arrows in a chat history list.
; This sometimes returns True in those lists otherwise.
if !util__isSpecialFocus() && getObjectTypeCode() != WT_Edit
	return False
endIf
return supportsEditCallbacks()
endFunction

int function caretVisible()
; Part of avoiding "blank" on up/down arrows in a chat history list.
; This returns True in those lists otherwise.
if isPCCursor() && !isVirtualPCCursor()
&& !util__isSpecialFocus() && getObjectTypeCode() != WT_Edit
	return False
endIf
return caretVisible()
endFunction

int function isTransparentWindow(handle hwnd)
if getWindowClass(hwnd) == "Chrome_RenderWidgetHostHWND"
	return False
endIf
return isTransparentWindow(hwnd)
endFunction

script moveToChatEntryField()
; Move focus to the chat input edit box for the currently shown conversation.
var object chatInput = sutil__getElement("uiaChatInput")
if !chatInput
	beep()
	return
elif !chatInput.isEnabled
|| chatInput.controlType == UIA_TextControlTypeId
	sayMessage(OT_Error, chatInput.Name)
elif chatInput.hasKeyboardFocus
	sayObjectTypeAndText()
else
	chatInput.getLegacyIAccessiblePattern().getIAccessible().accSelect(SelFlag_TakeFocus, 0)
endIf
endScript

script moveToChatHistoryList()
; Move focus to the displayed conversation's history list.
var object uiaTop = sutil__getElement("uiaTop")
if !uiaTop
	beep()
	return
endIf
; This finds the button immediately before the chat message "list" in tab order. Based on Skype 8.25.
; "List" is in quotes here because the chat messages have no container; they all descend from the same UIA element that contains the input box and other controls.
var string sCond = "name '%1' name '%2' or isOffscreen False and isKeyboardFocusable True and"
sCond = formatString(sCond, scCreateNewGroup, scAddToGroup)
var object btn = uia__findFirst(uiaTop, TreeScope_Subtree, sCond)
if !btn
	beep()
	return
endIf
speechOff()
btn.setFocus()
btn = btn.buildUpdatedCache()
var int safety = 10
while safety && !btn.hasKeyboardFocus
	pause()
	btn = btn.buildUpdatedCache()
	safety = safety -1
endWhile
pause()  ; allows any final events to go through silently for that focus change, before we cause the one we actually want.
speechOn()
if !btn.hasKeyboardFocus
	beep()
	return
endIf
typeKey("Tab")
endScript

script sayTypingStatus()
; :Synopsis Say who is typing
; This command identifies who is typing.
sutil__sayTypist()
endScript


;================================= Module sutil ================================
; Utility code for the Skype Electron scripts.
;
; Author:  Doug Lee

globals
; A cache of UIA elements and related information.
	collection sutil___cache

globals object sutil___walker
object function sutil__treeWalker(object el)
; Return a UIA tree walker set to the given element as its starting point.
; Returns the walker or null on error.
var object walker = sutil___walker
if !walker
	walker = uia__treeWalker("rawView")
	sutil___walker = walker
endIf
if !walker
	return
endIf
walker.currentElement = el
return walker
endFunction

prototype string function getUserName()
string function sutil__getUserID()
; Return an ID string suitable for identifying a single Skype Electron user.
; The string need not identify who the user is.
; Using computerName/userName because SkypeUserName is not available in Skype Electron.
; Note that computer names are often randomized by manufacturer, and user names are often generic, like Owner or User.
; Hopefully the combination of the two will better support things like counting script users than would IP addresses.
var string userName = getUserName()
var object o = util__createObject("wscript.network")
var string computername = o.computername
return formatString("%1/%2", computerName, userName)
endFunction

int function sutil__moveToElement(string id)
; Move focus to the UIA element with the given id (locally defined).
; Returns True on success and False on failure.
; This function speaks its own error messages and handles lists where it is necessary to focus the selected item rather than the list itself.
var object el = sutil__getElement(id)
if !sutil___isValid(el)
	sayMessage(OT_Error, sutil___msgFieldNotFound)
	return False
endIf
if !el.isKeyboardFocusable && el.controlType == UIA_ListControlTypeId
	; Lists in this app tend not to accept focus directly (el.isKeyboardFocusable is False).
	; There also tends to be no selected item, so just focus the first available element.
	; Note that this may not actually be the first element that can exist.
	var object el1 = uia__findFirst(el, TreeScope_Subtree, "isKeyboardFocusable True")
	if el1
		el = el1
	endIf
endIf
if el.hasKeyboardFocus
	sayObjectTypeAndText()
	return True
elif el.controlType != UIA_EditControlTypeId && el.setFocus()
	return True
endIf
; For some reason, at least the chat edit box focus setting works better from MSAA:
; Setting via UIA causes the next key typed into the box to move focus to the top-level window.
var object o = el.getLegacyIAccessiblePattern().getIAccessible()
if o
	o.accSelect(SelFlag_TakeFocus, 0)
	return True
elif el.setFocus()
	return True
endIf
sayMessage(OT_Error, sutil___msgNoFocus)
return False
endFunction

string function sutil__gotoSkypeScriptPage(string page, int deferred)
; "Runs" or returns the run command for the given page.
; Returns the null string if deferred is False and the command to run if deferred is True.
; page: The page to launch.
; deferred: False to run, True to return the command to be run as a string.
if deferred
	var string cmd = "sutil__gotoSkypeScriptPage(\"" +page +"\", 0)"
	return cmd
endIf

var string strBrailleDriver, string strBrailleDisplay, string strBraillePort
BrailleGetDeviceInfo(strBrailleDriver, strBrailleDisplay, strBraillePort)
var string strURLString = sutil___url +page
+"?JFWVersion="+IntToString(getJFWVersion())
+"&ScriptRevision=45"
+"&BrailleDisplay="+strBrailleDisplay
+"&SkypeProgramVersion=" +getVersionInfoString(getAppFilePath(), "FileVersion")
+"&WindowsVersion="+formatString("%1", util__getWindowsVersionInfo().osCode)
+"&UserID="+sutil__getUserID()
strURLString = stringReplaceSubstrings(strURLString, " ", "%20")

if UserBufferIsActive () UserBufferDeactivate() EndIf
SayFormattedMessage (OT_Message, sutil__msgSkypeScriptsHomePage_L, sutil__msgSkypeScriptsHomePage_S)
Run("\"" +strURLString +"\"")
endFunction

int function sutil___isPhoneNumber(string s)
; Returns True if the given string appears to be a phone number as might be found in Skype.
var
	int i, int n, string ch
s = stringTrimLeadingBlanks(stringTrimTrailingBlanks(s))
n = stringLength(s)
if n > 50
	; 50 is a pretty liberal limit for the length of a phone number...
	return False
endIf
i = 1
if stringLeft(s, 1) == "+"
	i = 2
endIf
while i <= n
	ch = substring(s, i, 1)
	if !stringIsBlank(ch) && !stringContains(",-.0123456789", ch)
		return False
	endIf
	i = i +1
endWhile
return True
endFunction

int function sutil__play(string sound)
; Play the indicated sound file.
var
	string spath
;spath = findJAWSSoundFile(sound)
spath = getJAWSSettingsDirectory() +"\\sounds\\" +sound
if !fileExists(spath)
	spath = getJAWSSettingsDirectory() +"\\" +sound
endIf
if !fileExists(spath)
	spath = findJAWSSoundFile(sound)
endIf
if spath
	playSound(spath)
endIf
endFunction

int function sutil___isValid(object el)
; Return True if el is valid and not obsolete.
if !el.controlType
	return False
endIf
; Obsolete objects in this app lose their parent references.
var object walker = sutil__treeWalker(el)
if !walker || !walker.gotoParent()
	return False
endIf
return True
endFunction

object function sutil__getElement(string id, optional int maxCacheAge)
; Get an element by a locally-defined id.
; Define a function for each id; for example, for the id uiaChatInput, define sutil___get_uiaChatInput().
var handle hwnd = getFocus()
if !sutil___cache.hwndTop || !sutil___isValid(sutil___cache.uiaTop)
|| sutil___cache.hwndTop != hwnd
	var object uiaTop = uia__elementFromWindow(hwnd)
	if !uiaTop
		return
	endIf
	sutil___cache = new collection
	sutil___cache.uiaTop = uiaTop
	sutil___cache.hwndTop = hwnd
endIf
var int tc = getTickCount()
var string tcname = id +"__tick"
var string tc0 = sutil___cache[tcname]
var object el = sutil___cache[id]
if (!maxCacheAge || tc-tc0 < maxCacheAge)
&& sutil___isValid(el)
	return el.buildUpdatedCache()
endIf
var string func = "sutil___get_" +id
el = callFunctionByName(func)
if !el
	; ToDo: Negative results not cached.
	return
endIf
sutil___cache[id] = el.buildUpdatedCache()
sutil___cache[tcname] = tc
return el
endFunction

void function sutil__resetCache()
; Reset the UIA element cache.
var variant null
sutil___cache = null
endFunction

object function sutil___get_uiaChatInput()
; Retrieve the chat input UIA element.
var object elTop = sutil___cache.uiaTop
if !sutil___isValid(elTop)
	return
endIf
var object walker = sutil__treeWalker(elTop)
if !walker
|| !walker.gotoNextSibling()
	return
endIf
var object el = walker.currentElement
walker = uia__treeWalker(formatString("controlType Edit controlType Text name '%1' and or controlType ListItem or", sutil___scNoMessaging))
if !walker
	return
endIf
walker.currentElement = el
if !walker.gotoPriorSibling()
	return
endIf
var int ctype = walker.currentElement.controlType
if ctype != UIA_EditControlTypeId && ctype != UIA_TextControlTypeId
	return
endIf
; ToDo: Could verify via prompt text but that would require localization.
return walker.currentElement
endFunction

object function sutil___get_uiaHistory()
; Retrieve the chat history UIA element.
; As of July 17, 2018 and based on Skype 8.25, this actually retrieves the latest chat message, as there is no container for them.
var object walker
var object el = sutil___cache.uiaChatInput
if !sutil___isValid(el)
	el = sutil___get_uiaChatInput()
endIf
if !sutil___isValid(el)
	return
endIf
walker = uia__treeWalker("controlType Pane isOffscreen False and isKeyboardFocusable True and")
if !walker
	return
endIf
walker.currentElement = el
if !walker.gotoPriorSibling()
	return
endIf
return walker.currentElement
endFunction

object function sutil___get_uiaConvName()
; Get an object that names the displayed conversation.
var object elConvList = sutil___cache.uiaConvList
if !sutil___isValid(elConvList)
	elConvList = sutil___get_uiaConvList()
endIf
if !sutil___isValid(elConvList)
	return
endIf
var object walker = uia__treeWalker("rawView")
if !walker
	return
endIf
walker.currentElement = elConvList
if !walker.gotoNextSibling()
	return
endIf
var object el = walker.currentElement
; If el is a button, it names the conversation. If it's not, it indicates the home screen is showing.
return el
endFunction

string function sutil__getConvName()
; Get a string identifying the conversation on screen, or that the home screen is showing instead.
var object el = sutil__getElement("uiaConvName")
if !el
	return ""
elif !el.name
	if el.controlType == UIA_ButtonControlTypeId
		; Unrecognized scenario: Button with description but no name. We don't know what to return here.
		return ""
	endIf
	; This fits the profile of the Home screen being displayed instead of a conversation.
	return sutil___msgHome
endIf
; The conversation/other-party name is in the MSAA Description property of this button.
var string name = el.getPropertyValue(UIA_LegacyIAccessibleDescriptionPropertyId)
if !name
	; Or it's supposed to be...
	return ""
endIf
; Chances are the Name property contains presence information. Useful in SayWindowTitle...
var string info = el.name
if stringStartsWith(info, name+", ")
	info = stringChopLeft(info, stringLength(name)+2)
	; Remaining format (in English but commas are assumed to mark divisions in all languages here): "profile, <presenceInfo>, View profile"
	; Strip off the first, then last section by commas without regard to presence info format or actual text content besides commas.
	sutil__string__segmentNext(info, ", ")
	sutil__string__segmentNextFromRight(info, ", ")
	name = formatString("%1, %2", name, info)
endIf
name = formatString(sutil___fmtConversation, name)
return name
endFunction

object function sutil___get_uiaConvList()
; Get the conversation list object.
var object elTop = sutil___cache.uiaTop
if !sutil___isValid(elTop)
	return
endIf
var object walker = uia__treeWalker("name '"+sutil___scConvList+"' isOffscreen False and")
if !walker
	return
endIf
walker.currentElement = elTop
if !walker.gotoFirstChild()
	return
endIf
return walker.currentElement
endFunction

int function sutil__sayTypist()
endFunction


;-------------------------------- Module string --------------------------------
; String manipulation functions.
;
; Author:  Doug Lee

string function sutil__string__segmentNext(string byRef s, string sep)
; Gets, returns, and removes from s the next string segment and the trailing separator if any.
; sep may be multi-character and is case sensitive.
; If sep is null, the entire string is treated as one segment.
var
	int pos,
	string s1
pos = 0
if sep
	pos = stringContains(s, sep)
endIf
if !pos || !sep
	s1 = stringChopLeft(" " +s, 1)  ; avoids pointer duplication
	s = ""
	return s1
endIf
s1 = stringLeft(s, pos-1)
s = stringChopLeft(s, pos +stringLength(sep) -1)
return s1
endFunction

string function sutil__string__segmentNextFromRight(string byRef s, string sep)
; Gets, returns, and removes from s the last string segment and the preceding separator if any.
; sep may be multi-character and is case sensitive.
; If sep is null, the entire string is treated as one segment.
var
	int pos,
	string s1
pos = 0
if sep
	pos = stringContainsFromRight(s, sep)
endIf
if !pos || !sep
	s1 = stringChopLeft(" " +s, 1)  ; avoids pointer duplication
	s = ""
	return s1
endIf
s1 = stringChopLeft(s, pos +stringLength(sep) -1)
s = stringLeft(s, pos-1)
return s1
endFunction

int function sutil__string__replace(string byRef s, string r1, string r2, string whatToDo)
; Replace r1 in s with r2 if r1 is found where whatToDo indicates.
; Note:  A space is sometimes prepended to and then removed from results in this function to keep JAWS from adding
; numeric strings arithmetically, which can happen in JAWS versions at least through 3.70.47 [DGL, 12/20/2000]
; whatToDo: l=left, r=right, a=anywhere, e=everywhere.
; Add t (for Trim) if whitespace should be removed at the point where s is split.
; Add c (for ConsiderCase) if case-sensitive matching is desired.
; For compatibility with older scripts, WhatToDo can also be an integer sum of flags:
; 0=l, 1=r, 2=a, 3=e, 8=t, 16=c.
; Returns the number of changes made to s, which must be passed by reference.
; If s is null, no changes are made and 0 is returned.
; If r1 is null, the following happens based on WhatToDo:
;	"e" - r2 is inserted between each character pair in s.  All blanks are removed first if "t" is included.
;	"l" or "a" - r2 is inserted after the first character of s, after blanks around it are removed if "t" is included.
;	"r" - r2 is inserted before the last character of s, after blanks around it are removed if "t" is included.
var
	string fromWhere,
	int trimEnds,
	int considerCase,
	int pos,
	int l1,
	string sa,
	string sb,
	int changed,
	int changeCount,
	int done

if !stringLength(s)
	return 0
endIf

; Compatibility with older integer whatToDo parameter.
if stringLength(whatToDo) == 1 && stringContains(whatToDo, "0")
	whatToDo = "l"
	elif stringToInt(whatToDo)
	whatToDo = stringSegment("r|a|e|||||lt|rt|at|et|||||lc|rc|ac|ec|||||ltc|rtc|atc|etc", "|", stringToInt(whatToDo) & 31)
endIf

whatToDo = stringLower(whatToDo)
trimEnds = stringContains(whatToDo, "t")
considerCase = stringContains(whatToDo, "c")
fromWhere = whatToDo

; Special cases for null r1.
if !stringLength(r1) && fromWhere == "e"
	if trimEnds
		sa = stringStripAllBlanks(s)
	else
		sa = sutil__string__ref(s)
	endIf
	s = ""
	l1 = stringLength(sa)
	pos = 1
	while pos < l1
		s = sutil__string__cat(s, sutil__string__cat(substring(sa, pos, 1), r2))
		pos = pos +1
	endWhile
	s = sutil__string__cat(s, stringRight(sa, 1))
	return stringLength(sa)
elif !stringLength(r1)
	if fromWhere == "l" || fromWhere == "a"
		if trimEnds
			sb = stringTrimLeadingBlanks(s)
		else
			sb = s +""
		endIf
		sa = stringLeft(sb, 1)
		sb = stringChopLeft(sb, 1)
		if trimEnds
			sb = stringTrimLeadingBlanks(sb)
		endIf
	else  ; fromWhere == "r"
		if trimEnds
			sb = stringTrimTrailingBlanks(s)
		else
			sb = s +""
		endIf
		sa = stringChopRight(sb, 1)
		if trimEnds
			sa = stringTrimTrailingBlanks(sa)
		endIf
		sb = stringRight(sb, 1)
	endIf
	if !stringLength(sa) || !stringLength(sb)
		return 0
	endIf
	s = sutil__string__cat(sa, sutil__string__cat(r2, sb))
	return 1
endIf

changeCount = 0
while !done
	l1 = stringLength(r1)
	changed = false
	if fromWhere == "l"
		if !stringCompare(stringLeft(s, l1), r1, considerCase)
			if trimEnds
				s = stringTrimLeadingBlanks(stringChopLeft(s, l1))
				s = r2 + s
			else
				s = r2 + stringChopLeft(s, l1)
			endIf
			changed = true
		endIf
	elif fromWhere == "r"
		if !stringCompare(stringRight(s, l1), r1, considerCase)
			if trimEnds
				s = stringTrimTrailingBlanks(stringChopRight(s, l1))
				s = s + r2
			else
				s = stringChopRight(s, l1) + r2
			endIf
			changed = true
		endIf
	else  ; fromWhere == "a" or "e"
		if considerCase
			pos = StringContains(s, r1)
		else
			pos = StringContains(stringLower(s), stringLower(r1))
		endIf
		if pos
			if trimEnds
				sa = stringTrimTrailingBlanks(stringLeft(s, pos-1))
				sb = stringTrimLeadingBlanks(stringChopLeft(s, pos+l1-1))
				s = " " +sa +r2 +sb
				s = stringChopLeft(s, 1)
			else
				s = " " +stringLeft(s, pos-1) +r2 +stringChopLeft(s, pos+l1-1)
				s = stringChopLeft(s, 1)
			endIf
			changed = True
		endIf
	endIf
	if fromWhere == "e"
		done = !changed
	else
		done = True
	endIf
	if changed
		changeCount = changeCount +1
	endIf
endWhile
return changeCount
endFunction

string function sutil__string__ref(variant v)
; Type cast from anything to string.
; Useful for allowing object results to be included in string concatenations without a compile-time error.
return v
endFunction

string function sutil__string__cat(string s1, string s2)
; Safe string concatenator.
return stringChopLeft(" " +s1 +s2, 1)
endFunction

;------------------------------ End module string ------------------------------

;=============================== End module sutil ==============================


;================================= Module mkcs =================================
; This module implements a KeymapChangedEvent-based multi-key command sequence system.
;
; Usage:
; Include this module in the main jss.
; Create a message constant for the MKCS help text (see the mkcs__help script for format info).
; Load the help text from AutoStartEvent with a call to mkcs__setHelpText().
; Set the command key (and again if it changes) by calling mkcs__setCommandKey().
; Write key mappings in the jkm file using "&" between layers on each line.
; For each command level, make the following additional mappings (the examples are for the top level with [ as the command key):
;	[&Escape=mkcs__exit
;	[&Tab*=mkcs__help("next")
;	[&Shift+Tab*=mkcs__help("prev")
;	[&JAWSKey+Tab*=mkcs__help("cur")
;	[&Insert+Tab*=mkcs__help("cur")
;	[&Enter=mkcs__help("exec")
; NOTE: The first line above is mandatory, as it is also used by this code
; to find the original command key when changing it to a new one.
; See mkcs__setCommandKey().
;
; Author:  Doug Lee
;
; Revision History:  See version control logs.

include "hjglobal.jsh"

; For support of JAWS versions before 14.0.
prototype string function getScriptSynopsis(string scriptName, optional int useDefault)
prototype string function getScriptDescription(string scriptName, optional int useDefault)
; And for JAWS versions before 17.0...
prototype string function getActiveConfiguration(optional int wantDomain)

globals
; The path to the sound file to play on exit from a layer.
	string mkcs___soundPath,
; For default-level scripts, the namespace for the CommandKey value in the MKCSOptions section of the effective config file.
	string mkcs___namespace,
; The file name and jcf section used by the scripts, if explicitly set in a mkcs__setCommandKey call.
	string mkcs___fname,
	string mkcs___sect,
; The command key that starts all sequences recognized here.
	string mkcs___commandKey

string function mkcs___getActiveConfiguration(optional int wantDomain)
; Wrapper for the JAWS getActiveConfiguration() function that returns the application config when True is passed but no domain scripts are active.
; The JAWS documentation says getActiveConfiguration does this, but through at least JAWS 18 it returns "" in this case. [DGL, 2017-08-19
if !wantDomain
	return getActiveConfiguration()
endIf
var string cfg = getActiveConfiguration(wantDomain)
if cfg
	return cfg
endIf
return getActiveConfiguration()
endFunction

script mkcs__editCommandKey()
; Allow the user to alter the command key used to start key sequences for these scripts.
; Traditionally assigned to JAWSKey+C / Insert+C at the first level after the currently active command key.
var string sKey = mkcs__getCommandKey()
var string sKey0 = sKey
var int okPressed = inputBox(mkcs___msgEditPrompt, mkcs___msgEditTitle, sKey)
if !okPressed || stringCompare(sKey, sKey0, True) == 0
	sayMessage(OT_User_Requested_Information, mkcs___msgNoChange)
	return
endIf

; Set the key in the active or appropriate config file first, so mkcs__setCommandKey() will find it correctly.
var string jcfname = mkcs___fname
; That would be a jkm file name if not blank.
if jcfname
jcfname = stringLeft(jcfname, stringContainsFromRight(jcfname, ".")-1) +".jcf"
else
	jcfname = mkcs___getActiveConfiguration(True) +".jcf"
endIf
if !stringContains(jcfname, "\\")
	jcfname = getJAWSSettingsDirectory() +"\\" +jcfname
endIf
var string jcfSect = "MKCSOptions"
var string jcfKey = mkcs___namespace +"CommandKey"
var string userCommandKey = iniReadString(jcfSect, jcfKey, "", jcfname)
if stringCompare(userCommandKey, sKey) == 0
	sayMessage(OT_User_Requested_Information, mkcs___msgNoChange)
	return
endIf
iniWriteString(jcfSect, jcfKey, sKey, jcfname)
mkcs__setCommandKey(sKey)
endScript

int function mkcs__setCommandKey(string sKey, /*?*/ string fname, string sect)
; Sets the key used as a prefix for all command sequences in this application.
; Also performs other minor initialization tasks.
; Call from AutoStartEvent or appropriate initialization code.
; Also call if the key is ever changed.
; Updates the jkm file if necessary.
; If fname is not passed, defaults to <activeConfiguration>.jkm.
; If sect is not passed, defaults to "Common Keys."
; The given command key can be overridden by a value for CommandKey in an MKCSOptions section of the config file as of June 13, 2017.
; When using this in default-level scripts, first call mkcs__setNamespace("MyScripts") so the value will come from MyScriptsCommandKey instead.
; The config file defaults to <activeConfiguration>.jcf but gets the basename from fname if passed.
; Returns 0 if things are not left in sync, 1 if they are, and 2 if the jkm changed.
; Things should not be left out of sync unless there is an error.
mkcs___soundPath = ""
if !sKey
	; Can't set to null key, doing nothing is inconsistent,
	; so we set to a string that is not a valid key.
	sKey = "<NoKey>"
endIf
mkcs___commandKey = sKey
if fname
	mkcs___fname = fname
else
	fname = mkcs___fname
endIf
if !fname
	fname = mkcs___getActiveConfiguration(True) +".jkm"
endIf
if !stringContains(fname, "\\")
	fname = getJAWSSettingsDirectory() +"\\" +fname
endIf
if sect
	mkcs___sect = sect
else
	sect = mkcs___sect
endIf
if !sect
	sect = "Common Keys"
endIf
; Allow an MKCSOptions config value to override the command key.
; This is mainly for non-English installations, where author command key choice may block standard things like accented letter typing.
var string jcfname = stringLeft(fname, stringContainsFromRight(fname, ".")-1) +".jcf"
var string jcfSect = "MKCSOptions"
var string jcfKey = mkcs___namespace +"CommandKey"
var string userCommandKey = iniReadString(jcfSect, jcfKey, "", jcfname)
if userCommandKey
	mkcs___commandKey = userCommandKey
endIf

var string jkm = util__stringFromFile(fname)
if !jkm
	; Chances are this means we're on the secure desktop and can't read files.
	return 0
endIf
var string sep = "\n"
var int pos = stringContains(stringLower(jkm), stringLower("&escape=mkcs__exit"))
if !pos
	sayMessage(OT_Error, mkcs___msgNotFound1)
	return 0
endIf
var string sKey0 = stringLeft(jkm, pos-1)
pos = stringContainsFromRight(sKey0, sep)
if !pos
	sayMessage(OT_Error, mkcs___msgNotFound2)
	return 0
endIf
sKey0 = stringChopLeft(sKey0, pos)
; In case the first layer entry is not the command key entry.
pos = stringContainsFromRight(sKey0, "&")
if pos
	sKey0 = stringLeft(sKey0, pos-1)
endIf
if stringCompare(sKey0, mkcs___commandKey, False) == 0
	return 1
endIf

sayUsingVoice(VCTX_Message, mkcs___msgUpdating, OT_Message)
var int n = stringSegmentCount(jkm, sep)
var string jkm1 = ""
var int i = 1
var int changed = False
var int inSect = False
var string line
while i <= n
	line = stringSegment(jkm, sep, i)
	line = stringTrimLeadingBlanks(stringTrimTrailingBlanks(line))
	if stringStartsWith(line, "[" +sect +"]")
		inSect = True
	elif stringLeft(line, 1) == "[" && stringRight(line, 1) == "]"
		inSect = False
	elif inSect
		if stringStartsWith(line, sKey0+"&")
			line = mkcs___commandKey +stringChopLeft(line, stringLength(sKey0))
			changed = True
		endIf
	endIf
	if jkm1
		jkm1 = jkm1 +"\r" +sep
	endIf
	jkm1 = jkm1 +line
	i = i +1
endWhile
jkm1 = jkm1 +"\r" +sep
if changed
	util__stringToFile(fname, jkm1)
	sayUsingVoice(VCTX_Message, mkcs___msgDone, OT_Message)
	scheduleFunction("reloadAllConfigs", 2)
	return 2
else
	sayMessage(OT_Error, mkcs___msgUpdateError)
	return 0
endIf
endFunction

string function mkcs__getCommandKey()
; Returns the key used as a prefix for all command sequences in this application.
; This will be the value passed to the latest call to mkcs__setCommandKey().
return mkcs___commandKey
endFunction

string function mkcs___getActiveKeyboardLayout()
; Returns the name of the keyboard layout in effect or null if this can't be determined.
; Tested in JAWS 9.0.2169.
var string fname, string layout
fname = getJAWSSettingsDirectory() +"\\default.jcf"
layout = iniReadString("Options", "KeyboardType", "", fname)
if !layout
	fname = getSharedSettingsDirectory() +"\\default.jcf"
	layout = iniReadString("Options", "KeyboardType", "", fname)
endIf
return layout
endFunction

string function mkcs___getActiveKeymapFile()
; Returns the full path to the active .jkm file or nul if this can't be determined.
var
	string fname
fname = mkcs___getActiveConfiguration(True)
if !fname
	return ""
endIf
fname = fname +".jkm"
fname = findJAWSSettingsFile(fname)
if !fname
	return ""
endIf
return fname
endFunction

int function mkcs___updateKeyList(string byRef list, string prefix, string jkm, string layout, optional int includeScriptNames)
; Adds any key sequences to the given list that begin with the given prefix.
; Key sequences in the given keyboard layout section of the indicated jkm file are scanned for matches.
; Returns the number (0 or more) of entries added to the list.
var
	string keys,
	string sKey, string sKey0,
	int pos,
	int n, int i,
	int prefixLen,
	int additions
keys = iniReadSectionKeys(layout +" Keys", jkm)
n = stringSegmentCount(keys, "|")
if !keys || !n
	return 0
endIf
prefix = stringStripAllBlanks(prefix)
prefixLen = stringLength(prefix)
additions = 0
i = 1
while i < n
	sKey = stringStripAllBlanks(stringSegment(keys, "|", i))
	sKey0 = sKey
	; Case-insensitive prefix match that doesn't let sKey be a substring of the prefix.
	if stringLeft(sKey, prefixLen) == prefix && stringLength(sKey) > prefixLen
		; +1 cuts off the "&" after the prefix.
		sKey = stringChopLeft(sKey, prefixLen+1)
		; Only consider this key level, not subsequent ones.
		; Also remove the trailing "*" that means "stay at this level" if it is there.
		pos = stringContains(sKey, "&")
		if pos
			sKey = stringLeft(sKey, pos-1)
		elif stringRight(sKey, 1) == "*"
			sKey = stringChopRight(sKey, 1)
		endIf
		; Avoid adding duplicates or keys that are part of this help system.
		if !stringContains(stringLower(sKey), "tab")
		&& !stringContains(stringLower(sKey), "enter")
		&& !stringContains(stringLower(sKey), "esc")
		&& !stringSegmentIndex(list, "|", sKey, True)
			list = list +"|" +sKey
			if includeScriptNames
				list = list +"=" +mkcs___getScriptAssignedTo(sKey0)
			endIf
			additions = additions +1
		endIf
	endIf
	i = i +1
endWhile
return additions
endFunction

string function mkcs___getKeyList(string prefix, optional int includeScriptNames)
; Returns a list of the key sequences starting with the given prefix.
; Key sequences in Common Keys and the active keyboard layout section of the active jkm file are scanned for matches.
; Returned sequences omit the prefix.
var
	string layout,
	string jkm,
	string list
layout = mkcs___getActiveKeyboardLayout()
if !layout
	return ""
endIf
jkm = mkcs___getActiveKeymapFile()
if !jkm
	return ""
endIf
list = ""
mkcs___updateKeyList(list, prefix, jkm, "Common", includeScriptNames)
mkcs___updateKeyList(list, prefix, jkm, layout, includeScriptNames)
list = stringChopLeft(list, 1)  ; remove extra leading separator.
return list
endFunction

const
; Values for iKeyStatus in keymapChangedEvent
	mkcs___KeySequenceComplete = 0,
	mkcs___KeySequenceRepeating = 1,
	mkcs___KeySequencePending = 2,
	mkcs___KeySequenceCanceled = 3,
	mkcs___KeySequenceStart = 4,
	mkcs___KeySequenceRestarted = 5
globals
	int mkcs___active,
	string mkcs___helpText,
	int mkcs___lastKeyCode,
	string mkcs___sequence,
; The next three are used for the Enter/exec functionality.
	string mkcs___lastSeq,
	string mkcs___execScript,
	int mkcs___lastIndex,
	int mkcs___lastKeyStatus,
	int mkcs___helpIndex

int function mkcs___sayKeyPrompt(string seq)
; Speak the prompt for the given (or default) key sequence if there is a prompt for it.
; Returns True if the speaking of the prompt was handled and False if not.
var
	int idx,
	string prompt,
	int isScheduledCall
isScheduledCall = False
if !seq
	seq = mkcs___sequence
	isScheduledCall = True
endIf
; Cut off the command key from the sequence before looking up prompts.
idx = stringContains(seq, "&")
if idx
	seq = stringChopLeft(seq, idx)
else
	seq = ""
endIf
; Find the key's entry in the help text.
idx = stringContains(stringLower(mkcs___helpText), "\10"+stringLower(seq)+"=")
if idx
	; Chop off everything through seq=
	prompt = stringChopLeft(mkcs___helpText, idx+stringLength(seq)+1)
	; Chop off everything past the key's entry.
	prompt = stringSegment(prompt, "\10", 1)
	; Chop off the actual help text (left of the prompt).
	prompt = stringSegment(prompt, "|", 2)
	if prompt
		if !isScheduledCall
			stopSpeech()
		endIf
		sayMessage(OT_Help, prompt)
	endIf
	return True
endIf
return False
endFunction

void function mkcs__setHelpText(string helpText)
; Set the help text to be used by this system.
; Call from AutoStartEvent in the application script file.
mkcs___helpText = "\10" +helpText +"\10"
endFunction

script mkcs__help(string which)
; Implements round-robin keyboard help for the current level of a multikey sequence.
; Which: next to move forward, prev to move backward, cur to stay put.
; These should correspond to Tab, Shift+Tab, and JAWSKey+Tab/Insert+Tab, respectively.
; Exec, which corresponds to Enter, executes the last entry reported.
; For all this to work, appropriate help must have been registered with mkcs__setHelpText().
; Required help format:
;	- One line per help entry.
;	- Line format: <sequence>=<help>[|<prompt>]
;	- <sequence> is the sequence to which the help applies, minus the command prefix and its "&".
;	- <help> is the help text to speak for this sequence. "@" means use the script's jsd Synopsis or, failing that, Description value.
;	- <prompt> (optional) is the prompt to speak when the user actually enters this sequence (not when getting help).
;	- The first line should provide the prompt for the actual command key (e.g., "=|Command").
var
	string prefix,
	int len,
	int pos,
	int i, int n,
	string seg,
	string helps,
	string seq

; Get the prefix, which is the current position in key sequences without the leading command key or its "&".
; Also get len, its length.
; Prefix will end with an "&" if it is not empty.
; It's empty for top-level keys though (keys that immediately follow the command key).
if which == "exec"
	; The sequence should have just now completed,
	; so the sequence is in mkcs___lastSeq, not mkcs___sequence.
	seq = mkcs___lastSeq
else
	seq = mkcs___sequence
endIf
prefix = seq
pos = stringContains(prefix, "&")
if pos
	; Get rid of the leading command character and its "&"
	; Then put an "&" at the end.
	prefix = stringChopLeft(prefix, pos) +"&"
else
	; We're at the top level (right after entry of the command key).
	prefix = ""
endIf
len = stringLength(prefix)

; Loop through help text lines, collecting those that apply to the current level.
; JAWS seems to use \10, not \13, to separate lines in a Messages text string.
n = stringSegmentCount(mkcs___helpText, "\10")
helps = ""
i = 2  ; mkcs__setHelpText prepends a delimiter, so the first segment is null.
while i <= n
	seg = stringSegment(mkcs___helpText, "\10", i)
	seg = stringTrimLeadingBlanks(stringTrimTrailingBlanks(seg))
	; The first check here keeps out the initial Command prompt line.
	; The second part lets through current entries if we're not at the top level.
	; The third lets through top-level entries if that's where we are.
	if stringLeft(seg, 1) != "=" && ((len && stringLeft(seg, stringLength(prefix)) == prefix) || !len)
		; Chop all but the current level's keystroke off this entry if we're not at the top level.
		if len
			seg = stringChopLeft(seg, len)
		endIf
		; Look for entries that are just this level, not deeper ones.
		; This logic allows "&" to occur in help or prompt text.
		if !stringContains(seg, "&") || stringContains(seg, "&") > stringContains(seg, "=")
			; Remove the prompt part if it's there.
			pos = stringContainsFromRight(seg, "|")
			if pos > stringContains(seg, "=")
				seg = stringLeft(seg, pos-1)
			endIf
			helps = helps +"|" +seg
		endIf
	endIf
	i = i +1
endWhile
helps = stringChopLeft(helps, 1)  ; extra leading delimiter

; Figure out which entry to speak, and do it.
if which == "next" || which == "prev"
	if which == "next"
		mkcs___helpIndex = mkcs___helpIndex +1
	elif which == "prev"
		mkcs___helpIndex = mkcs___helpIndex -1
	endIf
	if mkcs___helpIndex <= 0
		mkcs___helpIndex = stringSegmentCount(helps, "|")
	elif mkcs___helpIndex > stringSegmentCount(helps, "|")
		mkcs___helpIndex = 1
	endIf
endIf  ; next or prev
var int idx = mkcs___helpIndex
if which == "exec"
	idx = mkcs___lastIndex
endIf
seg = stringSegment(helps, "|", idx)
if !seg
	; Happens when there are no non-help-system keys defined.
	if which == "exec"
		beep()
	else
		sayMessage(OT_Error, mkcs___msgNoKeys)
	endIf
	return
endIf
if which == "exec"
	var string sKey = stringSegment(seg, "=", 1)
	;sayMessage(OT_User_Requested_Information, sKey)
	; TODO: Disabled as this does not work in a layer: The key is typed
	; but the script is not run. Simulating seq+"&"+sKey actually fails.
	; [DGL, 2014-08-24, JAWS 15.0]
	;simulateKey(sKey)
	if !mkcs___simulateKey(seq, sKey)
		beep()
	endIf
else
	; This says the help text, pauses, then says the keystroke.
	sKey = stringSegment(seg, "=", 1)
	var string txt = stringSegment(seg, "=", 2)
	if txt == "@"
		txt = mkcs___getScriptHelp(mkcs___getScriptAssignedTo(seq+"&"+sKey))
	endIf
	sayMessage(OT_User_Requested_Information, formatString("%2, %1",
		sKey, txt
	))
endIf
endScript

int function mkcs___simulateKey(string seq, string sKey)
; Simulate the final key in a multikey sequence by calling its script.
sKey = formatString("%1&%2", seq, sKey)
var string mapping = mkcs___getScriptAssignedTo(sKey)
if !mapping
	return False
endIf
; Don't call directly because this may cause recursion problems.
; Don't use scheduleFunction directly in case "()" appears in the jkm.
; Schedule a formatStringWithEmbeddedFunctions to handle all this.
mkcs___execScript = mapping
scheduleFunction("mkcs___exec", 0)
return True
endFunction

string function mkcs___getScriptAssignedTo(string sKey)
; Get the script name attached to the given key or sequence.
; Works on sequences where getScriptAssignedTo() fails.
var string layouts = mkcs___getActiveKeyboardLayout()
if layouts
	layouts = layouts +" Keys|"
endIf
layouts = layouts +"Common Keys"
var string files = ""
var string cfg = mkcs___getActiveConfiguration(True)
if cfg
	files = files +formatString("|%1|%2",
		getJAWSSettingsDirectory() +"\\" +cfg +".jkm",
		getSharedSettingsDirectory() +"\\" +cfg +".jkm"
	)
endIf
files = files +formatString("|%1|%2",
	getJAWSSettingsDirectory() +"\\default.jkm",
	getSharedSettingsDirectory() +"\\default.jkm"
)
files = stringChopLeft(files, 1)
var int nfiles = stringSegmentCount(files, "|")
var int nlayouts = stringSegmentCount(layouts, "|")
var int i, int j, string fpath, string layout
var string mapping = ""
for i = 1 to nfiles
	fpath = stringSegment(files, "|", i)
	for j = 1 to nlayouts
		layout = stringSegment(layouts, "|", j)
		mapping = iniReadString(layout, sKey, "", fpath)
		if mapping
			return mapping
		endIf
		mapping = iniReadString(layout, sKey+"*", "", fpath)
		if mapping
			return mapping
		endIf
	endFor
endFor
return ""
endFunction

void function mkcs___exec()
; Execute a script called via Enter.
formatStringWithEmbeddedFunctions("<$"+mkcs___execScript+">")
mkcs___execScript = ""
endFunction

int function mkcs___playExitSound()
; Play the "exit sequence" sound file.
if !mkcs___soundPath
	mkcs___soundPath = mkcs___findSound("mkcsExit.wav")
	if !mkcs___soundPath
		mkcs___soundPath = mkcs___findSound("CameraShutter3.wav")
	endIf
endIf
if mkcs___soundPath
	return playSound(mkcs___soundPath)
endIf
return False
endFunction

string function mkcs___findSound(string sound)
; Find and return the full path to the given sound file,
; or return null if it can't be found.
var string spath, string base
;spath = findJAWSSoundFile(sound)
base = getJAWSSettingsDirectory() +"\\"
spath = base +"sounds\\" +sound
if fileExists(spath)
	return spath
endIf
spath = getJAWSSettingsDirectory() +sound
if fileExists(spath)
	return spath
endIf
base = getSharedSettingsDirectory() +"\\"
spath = base +"sounds\\" +sound
if fileExists(spath)
	return spath
endIf
spath = getJAWSSettingsDirectory() +sound
if fileExists(spath)
	return spath
endIf
return ""
endFunction

void function keymapChangedEvent(int iKeyCode, string sKeyName, int iKeyStatus)
if mkcs__keymapChangedEvent(iKeyCode, sKeyName, iKeyStatus)
	return
endIf
if getJFWVersion() >= 1100000
	; Chaining in older JAWS versions causes "unknown function call to KeyMapChangedEvent" messages for some reason.
	; JAWS 11.0 starts including a KeymapChangedEvent in default.jss.
	keymapChangedEvent(iKeyCode, sKeyName, iKeyStatus)
endIf
endFunction

void function mkcs__keymapChangedEvent(int iKeyCode, string sKeyName, int iKeyStatus)
; Helper for KeyMapChangedEvent.
; Returns True if the default KeyMapChangedEvent should be blocked and False if not.

;sayString(formatString("Last %1 %2, current %4 %5",
;	mkcs___sequence, intToString(mkcs___lastKeyStatus), intToString(mkcs___lastKeyCode),
;sKeyName, intToString(iKeyStatus), intToString(iKeyCode)
;))
var
	string seq
; Filter out unrelated sequences and make any wanted sounds indicating changes into andout of our sequence.
if iKeyStatus == mkcs___KeySequenceStart || iKeyStatus == mkcs___KeySequenceRestarted
	if !isSameKey(sKeyName, mkcs___commandKey)
		; Not our sequence.
		mkcs___active = False
		return False
	endIf
	; Start or restart of our sequence.
	mkcs___active = True
elif !mkcs___active
	; Something other than activation going on, but it's not our sequence.
	return False
elif iKeyStatus == mkcs___KeySequenceCanceled
|| (iKeyStatus == mkcs___KeySequenceComplete && (mkcs___lastKeyStatus != mkcs___KeySequenceComplete && mkcs___lastKeyStatus != mkcs___KeySequenceCanceled))
	; End of our sequence.
	mkcs___playExitSound()
	mkcs___active = False
endIf

; Establish the sequence of keys that got us here.
if iKeyStatus == mkcs___KeySequenceStart || iKeyStatus == mkcs___KeySequenceRestarted
	mkcs___sequence = ""
	mkcs___helpIndex = 0
endIf
seq = ""
if mkcs___sequence
	seq = mkcs___sequence +"&"
endIf
seq = seq +sKeyName

; Announce the new key if there's an announcement to make for it.
if iKeyStatus == mkcs___KeySequenceCanceled
	if iKeyCode == mkcs___lastKeyCode
	&& (mkcs___lastKeyStatus == mkcs___KeySequenceStart || mkcs___lastKeyStatus == mkcs___KeySequenceRestarted)
		; Pass a repeat of the root key through and speak its name.
		sayMessage(OT_Keyboard, sKeyName)
		if !gbKeyboardHelp
			typeKey(sKeyName)
		endIf
	else
		; Avoid saying the invalid key's name and stop any previous speech.
		; Then announce the sequence cancellation.
		stopSpeech()
		;sayMessage(OT_Help, "Canceled", "")
	endIf
elif iKeyStatus != mkcs___KeySequenceComplete || mkcs___lastKeyStatus != mkcs___KeySequenceComplete
	; Say the key sequence's prompt if there is one, or the key name if not..
	; The above condition keeps this block from applying to non-layered script keys.
	if !mkcs___sayKeyPrompt(seq)
		stopSpeech()
		; TODO: Disabled because we don't always want default labels.
		;sayMessage(OT_Keyboard, sKeyName)
	endIf
endIf

; Set the sequence to remember for next time.
; Also set other globals to remember parameters to this call.
if iKeyStatus == mkcs___KeySequenceComplete || iKeyStatus == mkcs___KeySequenceCanceled
	if iKeyStatus == mkcs___KeySequenceComplete
		mkcs___lastSeq = mkcs___sequence
		mkcs___lastIndex = mkcs___helpIndex
	endIf
	mkcs___sequence = ""
	mkcs___helpIndex = 0
elif iKeyStatus == mkcs___KeySequenceRepeating
	; Do nothing; the new key stayed at this level.
	;scheduleFunction("mkcs___sayKeyPrompt", 0)
else
	; Pending:  Include the new key.
	; Start/Restarted:  Start with only the new key (already done above).
	mkcs___sequence = seq
	mkcs___helpIndex = 0
endIf
mkcs___lastKeyCode = iKeyCode
mkcs___lastKeyStatus = iKeyStatus
; This avoids an extra top-level prompt on the first non-Command scripted key after a sequence cancellation.
if iKeyStatus == mkcs___KeySequenceCanceled
	mkcs___lastKeyStatus = mkcs___KeySequenceComplete
endIf

; Chain this event.
;sayString("map " +sKeyName +" " +stringSegment("complete|repeating|pending|canceled|start|restarted", "|", iKeyStatus+1))
return True
endFunction

void function cancelLayeredKeySequence()
if mkcs___active
	mkcs___playExitSound()
	mkcs___active = False
endIf
cancelLayeredKeySequence()
endFunction

script mkcs__exit()
;sayMessage(OT_Help, "Exit", "", False)
endScript

int function mkcs__isSequenceActive()
; Returns True if one of our key sequences is active (in progress).
return mkcs___active
endFunction

string function mkcs___getScriptHelp(string scriptName)
; Get help for the given script. Tries Synopsis, Description, and default Synopsis and Description, in that order.
; On failure, returns the null string.
var string h = getScriptSynopsis(scriptName)
if !h h = getScriptDescription(scriptName) endIf
if !h h = getScriptSynopsis(scriptName, True) endIf
if !h h = getScriptDescription(scriptName, True) endIf
return h
endFunction

;=============================== End module mkcs ===============================


;================================ Module apphelp ===============================
; Automatic access to custom application help (HTM) file.
;
; The help file should be named <configName>.htm, <configName> being the result of getActiveConfiguration().
; Alternatively, the file can be named by a HelpDocument key in a Script Options section of the active configuration's .jcf file.
; As a final alternative, the name can be supplied from autoStartEvent by calling apphelp__setDocName().
;
; Author:  Doug Lee
;
; This is public code.

globals
; Custom doc name if set by a call to apphelp__setDocName().
	string apphelp___docName,
; App in focus when the above name was set.
	string apphelp___docApp,
; Number of quick-succession calls to screenSensitiveHelp seen so far.
; apphelp___doHelp() also sets this to -1 to tell the screenSensitiveHelp script in this file to bypass itself and call up the chain.
	int apphelp___callCount

string function apphelp___getCustomHelpDoc()
; Gets the full path to the custom help document for the running scripts.
; Returns null if no such file is found.
var string docname = ""
var string appname = getActiveConfiguration()
if stringLength(apphelp___docApp)
&& stringCompare(appname, apphelp___docApp, True) == 0
	docname = apphelp___docName
endIf
if !docName
	var string fname = getActiveConfiguration() +".jcf"
	fname = findJAWSSettingsFile(fname)
	if stringLength(fname) && fileExists(fname)
		docname = iniReadString("Script Options", "HelpDocument", "", fname)
	endIf
endIf
if !docname
	docname = getActiveConfiguration()
	if !stringLength(docname) || docname == "default"
		return ""
	endIf
	docname = docname +".htm"
endIf
docname = findJAWSSettingsFile(docname)
if !fileExists(docname)
	return ""
endIf
return docname
endFunction

void function apphelp___doHelp()
; Implements the following behavior based on the number of JAWSKey+F1 presses and the presence or absence of a custom (.htm) help file:
;	One press: Normal ScreenSensitiveHelp in 0.6 seconds.
;	Two quick presses: The custom help document opens if present, or if not, JAWS app-specific (chm) help opens.
;	Three or more quick presses: JAWS-specific (chm) help opens if present.
var
	int callCount,
		string docname
callCount = apphelp___callCount
apphelp___callCount = 0
if callCount != 2
	; 1 = screenSensitiveHelp, 3+ = app-specific help.
	; The standard screenSensitiveHelp script will use isSameScript() to decide what to do.
	apphelp___callCount = -1
	performScript screenSensitiveHelp()
	apphelp___callCount = 0
	return
endIf

docname = apphelp___getCustomHelpDoc()
if docname
	if userBufferIsActive()
		userBufferDeactivate()
	endIf
	sayMessage(OT_Message, apphelp___msgOpening)
	run("\"" +docname +"\"")
	return
endIf

; No custom help document, so revert to standard behavior.
apphelp___callCount = -1
performScript screenSensitiveHelp()
apphelp___callCount = 0
endFunction

script screenSensitiveHelp()
; Intercept that makes the apphelp___doHelp() function take effect.
if apphelp___callCount == -1
	; apphelp___doHelp sets apphelp___callCount to -1 to indicate it wants to call the original screenSensitiveHelp script.
	performScript screenSensitiveHelp()
	return
endIf

apphelp___callCount = apphelp___callCount +1
scheduleFunction("apphelp___doHelp", 6)
endScript

void function apphelp__setDocName(string docname)
; Set the name of the custom help doc to load.
; Call from AutoStartEvent if necessary.
apphelp___docApp = getActiveConfiguration()
apphelp___docname = docname
endFunction

;============================== End module apphelp =============================


;================================== Module uia =================================
; Code for accessing and manipulating the UIAScriptAPI in JAWS 14 and later.
;
; Globals not under namespace control: gc_cachedAPIObjects.
;
; Author: Doug Lee of Level Access

globals
; The main UIA interface, cached in case this speeds things.
	object uia___gUIA,
; An optional cache of API objects.
; This is a workaround for a JAWS (at least 14-18) issue whereby retrieval of an API object can start failing until the next JAWS restart.
; This global is deliberately JAWS-wide, not namespace-controlled, so that one script set building a cache will benefit all others.
	collection gc_cachedAPIObjects,
; An indicator of the version of the UIA support (see uia__version()).
; Set by the first uia__getScriptAPI() call.
	int uia___version,
; Lists of IDs from UIA.jsh / UI Automation.
	collection uia___props,
	collection uia___vals,
	collection uia___events,
	collection uia___attributes,
	collection uia___textUnits,
	collection uia___fullnames,
; A flag indicating when all those lists are completely built.
	int uia___built

; Messages for this code that should not be localized,
; because they are for scripters and should not be heard by users.
messages
@uia___msgInvalidScopeChar
Invalid search scope character: %1
@@
@uia___msgNullConditionString
Null condition string
@@
@uia___msgStringNotTerminated
String literal in condition spec was not terminated
@@
; %1 is a reference like %11 (only up to %10 allowed).
@uia___msgInvalidRef
Invalid reference in condition: %1
@@
; %1 is a list type (property, value, event, etc.).
; %2 is an invalid property/value/event/etc. name.
@uia___msgInvalidName
Module uia: Invalid %1 name %2
@@
; %1 is the number of unused stack values.
@uia___msgSpecIncomplete
%1 value(s) not used - missing trailing 'and'?
@@
@uia___msgUnknownError
Unknown error in condition specification
@@
endMessages

object function uia__getScriptAPI(optional int forceNew)
; Get the main UIAScriptAPI object.
; Also sets uia__version on first call.
; If forceNew is True, the object is a new reference.
; Otherwise it may (but may not) be a shared object.
; Shared objects may be shared by other scripts.
; You may want to pass True when hooking events.
if !uia___gUIA
	uia___gUIA = uia___rawCreate()
	if !uia___gUIA
		; Use the API object cache if somebody built one.
		uia___gUIA = uia___getCachedAPIObject(forceNew)
	endIf
	if !uia___gUIA
		return
	endIf
	if !uia___version
		; uia__version() sets uia___version to avoid recursion.
		uia___version = uia__version(uia___gUIA)
	endIf
endIf
scheduleFunction("uia___UIAClear", 50)
if forceNew
	var object newObj = uia___rawCreate()
	if !newObj
		; Use the API object cache if somebody built one.
		newObj = uia___getCachedAPIObject(forceNew)
	endIf
	return newObj
endIf
return uia___gUIA
endFunction

object function uia___rawCreate()
; Raw creator of the UIAScriptAPI object.
var object uia = createObjectEx("FreedomSci.UIA", FALSE, "UIAScriptAPI.x.manifest")
return uia
endFunction

int function uia__version(optional object oUIA)
; Indicate the version of the available or given UIA support:
;	-1 - Unknown.
;	0 - Not set yet.
;	1 - JAWS 14.
;	2 - JAWS 15 before update 1.
;	3 - JAWS 15 update 1 (15.0.5056).
; The version of the available UIA support is cached by uia__getScriptAPI().
var int ver = -1
if !oUIA
	if uia___version
		; Already figured out for this JAWS instance.
		return uia___version
	endIf
	; uia__getScriptAPI() passes a valid object,
	; so this will not cause recursion.
	; Set uia___version here to keep uia__getScriptAPI() from calling us though.
	uia___version = -1
	oUIA = uia__getScriptAPI()
	if !oUIA
		return -1
	endIf
endIf
var variant names = COMGetAvailableNames(oUIA)
names = stringLower(names)
if stringContains(names, "addtextedittextchangedeventhandler()")
	; Added in JAWS 15.0.5056.
	; Anything newer than that JAWS version must be detected here.
	ver = 3
elif stringContains(names, "events()")
	; Included in JAWS 14 but removed in JAWS 15.
	ver = 1
else
	; All we have left is the initial JAWS 15 release.
	ver = 2
endIf
return ver
endFunction

object function uia___getCachedAPIObject(int forceNew)
; Return (and remove from cache if not shared) a cached UIAScriptAPI object.
var collection cache = gc_cachedAPIObjects
if !cache
	return
endIf
var int i = 1
if forceNew
	i = cache.count
	if i <= 1
		; Out of new ones... (cache item 1 is reserved as the shared object).
		uia___reportAPICacheSize()
		return
	endIf
endIf
var object api = cache[intToString(i)]
if i > 1
	var variant null
	cache[intToString(i)] = null
	cache.count = cache.count -1
endIf
uia___reportAPICacheSize()
return api
endFunction

void function uia___reportAPICacheSize()
; Indicate to the user the API object cache size. This is called when the size changes.
var string fname = getJAWSSettingsDirectory() +"\\Sounds\\gauge" +intToString(gc_cachedAPIObjects.count) +".wav"
if !fileExists(fname)
	return
endIf
playSound(fname)
endFunction

object function uia__elementFromWindow(handle hwnd)
; Return the UIA element rooted at the given window or null on error.
var object oUIA = uia__getScriptAPI()
if !oUIA
	return
endIf
return oUIA.getElementFromHandle(hwnd)
endFunction

object function uia__treeWalker(variant cond, optional
	variant v1, variant v2, variant v3, variant v4, variant v5,
	variant v6, variant v7, variant v8, variant v9, variant v10
)
; Create and return a treeWalker using the given condition.
; The condition may be an object as expected by oUIA.CreateTreeWalker(),
; or a string as expected by uia__condition().
var object oUIA = uia__getScriptAPI()
if !oUIA
	return
endIf
if getVariantType(cond) == 8  ; VT_String
	var int ver = uia__version()
	if ver >= 3 && stringCompare(cond, "rawView", False) == 0
		return oUIA.RawViewWalker
	elif ver >= 3 && stringCompare(cond, "contentView", False) == 0
		return oUIA.contentViewWalker
	elif ver >= 3 && stringCompare(cond, "controlView", False) == 0
		return oUIA.controlViewWalker
	else
		cond = uia__condition(cond, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10)
	endIf
	if !cond
		return
	endIf
endIf
return oUIA.CreateTreeWalker(cond)
endFunction

object function uia__findFirst(object el, int treeScope, variant cond, optional
	variant v1, variant v2, variant v3, variant v4, variant v5,
	variant v6, variant v7, variant v8, variant v9, variant v10
)
; Return the first match to the given condition.
; The condition may be an object as expected by oUIA.FindFirst(),
; or a string as expected by uia__condition().
if getVariantType(cond) == 8  ; VT_String
	cond = uia__condition(cond, v1, v2, v3, v4, v5, v6, v7, v8, v9, v10)
	if !cond
		return
	endIf
endIf
return el.findFirst(treeScope, cond)
endFunction

; Cache of conditions by spec, excluding those with %<n> parameters.
globals collection uia___conds

object function uia__condition(string spec, optional
	variant v1, variant v2, variant v3, variant v4, variant v5,
	variant v6, variant v7, variant v8, variant v9, variant v10
)
; Build and return a UIAScriptAPI Condition object from a string specification.
; Values and other Condition objects may be passed and referred to in the spec.
; The spec is in post-fix notation, case-insensitive.
; Items in the spec are separated by a single space or line ending.  Allowed items:
;	contentView, controlView, rawView: Require the given UIA tree view type.
;	focusedProcess: Require the process ID to match the focused application's process ID.
;	FPRaw/Control/ContentView: One of the above with focusedProcess.
;	not: Reverse the previous condition: "focusedProcess not."
;	and, or: Combine previous two conditions: "contentView focusedProcess and."
;	property name: The name of a property with a PropertyID.
;		These are from UIA.jsh without UIA_ prefix and PropertyId suffix.
;		Examples: Name, ControlType.
;	value name: The name of a value with an ID. Categories:
;		Control types like CheckBox, Edit, and Hyperlink.
;		Other constants without *_ prefix, like Horizontal and LargeIncrement.
;		Other raw constants like Next that contain no underscore.
;	False, True: Create a boolPropertyCondition with the previous item.
;		Example: "IsEnabled True."
;		If not after a property, creates a False or True condition
;		(directly in JAWS 15.0.5056+, simulated before that).
;	int like 1, -50, etc.: An integer value.
;		Makes an IntPropertyCondition with the previous item.
;		Example: "RangeValueMinimum -5"
;	String like 'hello': A string value.
;		Makes a StringPropertyCondition with the previous item.
;		Example: "Name 'My UIA Application'"
;		Apostrophes in the string are written as ''. Backslash escapes like \t also work.
;	%n (1-10): Use argument n, 1 being the first after the spec, etc.
;	May be a string or a Condition object. Ints are not supported here.
; Examples:
;	c1 = uia__condition("ControlType GroupControl Name %1 and", getObjectName(True, 1))
;		(Must be a group control named whatever getObjectName(True, 1) returns.)
;	c2 = uia__condition("ContentView %1 and", c1)
;		Same as c1 but must also be in the ContentView tree.
; Returns an IDispatch Condition object for further use with the UIAScriptAPI.
; Returns null on error and may speak a message explaining the problem.
if !uia___conds
	uia___conds = new collection
endIf
var object cnd, variant null
spec = stringReplaceChars(spec, "\10\13\160", " ")
spec = stringReplaceSubstrings(spec, "  ", " ")
if !stringContains(spec, "%")
	cnd = uia___conds[spec]
	if cnd
		return cnd
	endIf
endIf
var string spec0 = spec
var object oUIA = uia__getScriptAPI()
if !oUIA
	uia___sayError(uia___msgNoUIAObject)
	return
endIf
spec = stringTrimLeadingBlanks(stringTrimTrailingBlanks(spec))
if !spec
	uia___sayError(uia___msgNullConditionString)
	return
endIf
; Shortcuts allowed starting in JAWS 15.0.5056.
if uia__version() >= 3
	cnd = null
	if stringCompare(spec, "rawview", False) == 0
		cnd = oUIA.rawViewCondition
	elif stringCompare(spec, "controlview", False) == 0
		cnd = oUIA.controlViewCondition
	elif stringCompare(spec, "contentview", False) == 0
		cnd = oUIA.contentViewCondition
	endIf
	if cnd
		uia___conds[spec0] = cnd
		return cnd
	endIf
endIf
var variant stack = uia__stl__stack(stringLength(spec))
var string sItem, int pos
var object c1, object c2
uia___buildLists()
; This is used to disambiguate %1 etc. as condition versus value.
var int expectValue = False
while spec
	if stringLeft(spec, 1) == "'"
		; A string literal that may contain '' apostrophe escapes.
		; JAWS will translate the other escapes like \t, \n, etc.
		expectValue = False
		spec = stringChopLeft(spec, 1)
		sItem = ""
		pos = stringContains(spec, "'")
		if !pos
			uia___sayError(uia___msgStringNotTerminated, spec0)
			return
		endIf
		while substring(spec, pos+1, 1) == "'"
			; Include the first of the two apostrophes.
			sItem = formatString("%1%2", sItem, stringLeft(spec, pos))
			; Then chop both off the spec.
			spec = stringChopLeft(spec, pos+1)
			; And finally find the next one.
			pos = stringContains(spec, "'")
			if !pos
				uia___sayError(uia___msgStringNotTerminated, spec0)
				return
			endIf
		endWhile
		sItem = formatString("%1%2", sItem, stringLeft(spec, pos-1))
		spec = stringChopLeft(spec, pos)
		uia__stl__push(stack, oUIA.CreateStringPropertyCondition(
			uia__stl__pop(stack).value, sItem
		))
	else
		pos = stringContains(spec, " ")
		if pos
			sItem = stringLeft(spec, pos-1)
			spec = stringChopLeft(spec, pos)
		else
			sItem = spec
			spec = ""
		endIf
		; Some shortcuts for developer (and slight runtime) efficiency.
		if stringStartsWith(sItem, "fp") && stringContains("fprawview fpcontrolview fpcontentview", stringLower(sItem))
			var object c0
			if sItem == "FPRawView"
				c0 = oUIA.CreateRawViewCondition()
			elif sItem == "FPControlView"
				c0 = oUIA.CreateControlViewCondition()
			else  ; FPContentView
				c0 = oUIA.CreateContentViewCondition()
			endIf
			c0 = oUIA.CreateAndCondition(c0,
				oUIA.CreateIntPropertyCondition(uia__propertyID("processID"),
				oUIA.GetFocusedElement().ProcessID)
			)
			uia__stl__push(stack, c0)
			expectValue = False
		elif stringCompare(sItem, "RawView", False) == 0
			uia__stl__push(stack, oUIA.CreateRawViewCondition())
			expectValue = False
		elif stringCompare(sItem, "ControlView", False) == 0
			uia__stl__push(stack, oUIA.CreateControlViewCondition())
			expectValue = False
		elif stringCompare(sItem, "ContentView", False) == 0
			uia__stl__push(stack, oUIA.CreateContentViewCondition())
			expectValue = False
		elif stringCompare(sItem, "FocusedProcess", False) == 0
			uia__stl__push(stack,
				oUIA.CreateIntPropertyCondition(uia__propertyID("processID"),
				oUIA.GetFocusedElement().ProcessID)
			)
			expectValue = False
		elif stringCompare(sItem, "True", False) == 0
			uia___handleTrueAndFalse(stack, oUIA, True)
			expectValue = False
		elif stringCompare(sItem, "False", False) == 0
			uia___handleTrueAndFalse(stack, oUIA, False)
			expectValue = False
		elif stringCompare(sItem, "not", False) == 0
			uia__stl__push(stack,
				oUIA.CreateNotCondition(uia__stl__pop(stack).value)
			)
			expectValue = False
		elif stringCompare(sItem, "and", False) == 0
			c2 = uia__stl__pop(stack).value
			c1 = uia__stl__pop(stack).value
			uia__stl__push(stack, oUIA.CreateAndCondition(c1, c2))
			expectValue = False
		elif stringCompare(sItem, "or", False) == 0
			c2 = uia__stl__pop(stack).value
			c1 = uia__stl__pop(stack).value
			uia__stl__push(stack, oUIA.CreateOrCondition(c1, c2))
			expectValue = False
		elif stringToInt(sItem) || sItem == "0" || sItem == "-0"
				uia__stl__push(stack,
					oUIA.CreateIntPropertyCondition(uia__stl__pop(stack).value, stringToInt(sItem))
				)
			expectValue = False
		elif stringLeft(sItem, 1) == "%"
			var int idx = stringToInt(stringChopLeft(sItem, 1))
			var variant v
			if idx == 1 v = v1
			elif idx == 2 v = v2
			elif idx == 3 v = v3
			elif idx == 4 v = v4
			elif idx == 5 v = v5
			elif idx == 6 v = v6
			elif idx == 7 v = v7
			elif idx == 8 v = v8
			elif idx == 9 v = v9
			elif idx == 10 v = v10
			else
				uia___sayError(formatString(uia___msgInvalidRef, sItem), spec0)
				return
			endIf
			if expectValue
				; ToDo: Assuming references like this are strings.
				uia__stl__push(stack, oUIA.CreateStringPropertyCondition(
					uia__stl__pop(stack).value, v
				))
			else
				uia__stl__push(stack, v)
			endIf
			expectValue = False
		else
			; All that's left are property and value names.
			var int val
			if collectionItemExists(uia___props, sItem)
				; Save for the value to compare it to.
				val = uia__propertyID(sItem)
				uia__stl__push(stack, val)
				expectValue = True
			else
				val = uia__valueID(sItem)
				if val == uia__Error
					return
				endIf
				; Now we have one of those values.
				; All of these should be ints.
				uia__stl__push(stack, oUIA.CreateIntPropertyCondition(
					uia__stl__pop(stack).value, val
				))
				expectValue = False
			endIf  ; Property name or value name.
		endIf  ; list of possible keywords.
	endIf  ; leading ' or no leading '.
	spec = stringTrimLeadingBlanks(spec)
endWhile
if uia__stl__size(stack) != 1
	uia___sayError(formatString(uia___msgSpecIncomplete, intToString(uia__stl__size(stack)-1)), spec0)
	return
endIf
var object condition = uia__stl__pop(stack).value
if !condition
	uia___sayError(uia___msgUnknownError, spec0)
endIf
uia___conds[spec0] = condition
return condition
endFunction

void function uia___handleTrueAndFalse(variant stack, object oUIA, int val)
; Handle "false" and "true" in specs.
; Helper for uia__condition().
var int standAlone = False
var variant prevVal
var int size = uia__stl__size(stack)
if size < 1
	standAlone = True
else
	prevVal = uia__stl__pop(stack).value
	if getVariantType(prevVal) == 9  ; VT_Dispatch
		; This should be a Condition object.
		standAlone = True
	endIf
endIf
if standAlone
	var object newVal
	if val
		if uia__version() >= 3
			newVal = oUIA.CreateTrueCondition()
		else
			; Something that should always be True.
			newVal = oUIA.CreateRawViewCondition()
		endIf
	else
		if uia__version() >= 3
			newVal = oUIA.CreateFalseCondition()
		else
			; Something that should always be False.
			; This is controlType = <invalidValue>.
			newVal = oUIA.CreateIntPropertyCondition(30003, 0xFFF00000L)
		endIf
	endIf
	uia__stl__push(stack, newVal)
	return
endIf
uia__stl__push(stack,
	oUIA.CreateBoolPropertyCondition(prevVal, val)
)
endFunction

variant function uia__propertyValue(object el, variant id)
; Get a UIA property value by its ID or name.
; Returns null and speaks an error message if an invalid property name is passed.
if !stringToInt(id)
	id = uia__propertyId(id)
	if id == uia__Error
		return
	endIf
endIf
return el.getPropertyValue(id)
endFunction

int function uia__propertyID(string name, optional int quiet)
; Get a UIA property ID.
; Returns the constant uia__Error if an invalid property name is passed.
; Also speaks an error message if that happens unless quiet is True.
return uia___getID("property", name, quiet)
endFunction

int function uia__valueID(string name)
; Get a UIA value ID.
; Returns the constant uia__Error if an invalid value name is passed.
; Also speaks an error message if that happens.
return uia___getID("value", name)
endFunction

int function uia__eventID(string name)
; Get a UIA event ID.
; Returns the constant uia__Error if an invalid event name is passed.
; Also speaks an error message if that happens.
return uia___getID("event", name)
endFunction

int function uia__attributeID(string name)
; Get a UIA attribute ID.
; Returns the constant uia__Error if an invalid attribute name is passed.
; Also speaks an error message if that happens.
return uia___getID("attribute", name)
endFunction

int function uia__textUnitID(string name)
; Get a UIA text unit id.
; Returns the constant uia__Error if an invalid text unit name is passed.
; Also speaks an error message if that happens.
return uia___getID("textUnit", name)
endFunction

collection function uia__getList(string listType)
; Get a list by a string identifier.
; The list is returned as a collection, name to value.
uia___buildLists()
var collection lst
if listType == "property"
	lst = uia___props
elif listType == "value"
	lst = uia___vals
elif listType == "event"
	lst = uia___events
elif listType == "attribute"
	lst = uia___attributes
elif listType == "textUnit"
	lst = uia___textUnits
endIf
return lst
endFunction

int function uia___getID(string listType, string name, optional int quiet)
; Get a property, value, event, attribute, or text unit id.
; Returns the constant uia__Error if an invalid name is passed.
; Also speaks an error message if that happens unless quiet is True.
var int result = uia___getID0(listType, name, quiet)
if result != uia__Error
	return result
endIf
; Rebuild lists in case they were updated since the last JAWS restart.
; (This implies a script update and reload.)
uia___buildLists(True)
result = uia___getID0(listType, name, quiet)
if result != uia__Error
	return result
endIf
if !quiet
	uia___sayError(formatString(uia___msgInvalidName, listType, name))
endIf
return uia__Error
endFunction

int function uia___getID0(string listType, string name, optional int quiet)
; Helper for uia___getID().
var collection lst = uia__getList(listType)
if stringLeft(name, 4) == "UIA_"
	name = stringChopLeft(name, 4)
endIf
if collectionItemExists(lst, name)
	return lst[name]
elif collectionItemExists(uia___fullnames, name)
	return uia___fullnames[name]
endIf
return uia__Error
endFunction

void function uia___UIAClear()
; Release the UIAScriptAPI cached object.
var variant null
uia___gUIA = null
endFunction

void function uia___buildLists(optional int force)
; Build internal lists of property, value, event, attribute, and textUnit Ids.
if uia___built && !force
	return
endIf
uia___built = False
uia___props = new collection
uia___vals = new collection
uia___events = new collection
uia___attributes = new collection
uia___textUnits = new collection
uia___fullnames = new collection
uia___fullnames["UIATrue"] = 0xffffffff
uia___vals["UIATrue"] = 0xffffffff
uia___fullnames["UIAError"] = 0xfffe
uia___vals["UIAError"] = 0xfffe
uia___fullnames["ButtonControlTypeId"] = 50000
uia___vals["Button"] = 50000
uia___fullnames["CalendarControlTypeId"] = 50001
uia___vals["Calendar"] = 50001
uia___fullnames["CheckBoxControlTypeId"] = 50002
uia___vals["CheckBox"] = 50002
uia___fullnames["ComboBoxControlTypeId"] = 50003
uia___vals["ComboBox"] = 50003
uia___fullnames["EditControlTypeId"] = 50004
uia___vals["Edit"] = 50004
uia___fullnames["HyperlinkControlTypeId"] = 50005
uia___vals["Hyperlink"] = 50005
uia___fullnames["ImageControlTypeId"] = 50006
uia___vals["Image"] = 50006
uia___fullnames["ListItemControlTypeId"] = 50007
uia___vals["ListItem"] = 50007
uia___fullnames["ListControlTypeId"] = 50008
uia___vals["List"] = 50008
uia___fullnames["MenuControlTypeId"] = 50009
uia___vals["Menu"] = 50009
uia___fullnames["MenuBarControlTypeId"] = 50010
uia___vals["MenuBar"] = 50010
uia___fullnames["MenuItemControlTypeId"] = 50011
uia___vals["MenuItem"] = 50011
uia___fullnames["ProgressBarControlTypeId"] = 50012
uia___vals["ProgressBar"] = 50012
uia___fullnames["RadioButtonControlTypeId"] = 50013
uia___vals["RadioButton"] = 50013
uia___fullnames["ScrollBarControlTypeId"] = 50014
uia___vals["ScrollBar"] = 50014
uia___fullnames["SliderControlTypeId"] = 50015
uia___vals["Slider"] = 50015
uia___fullnames["SpinnerControlTypeId"] = 50016
uia___vals["Spinner"] = 50016
uia___fullnames["StatusBarControlTypeId"] = 50017
uia___vals["StatusBar"] = 50017
uia___fullnames["TabControlTypeId"] = 50018
uia___vals["Tab"] = 50018
uia___fullnames["TabItemControlTypeId"] = 50019
uia___vals["TabItem"] = 50019
uia___fullnames["TextControlTypeId"] = 50020
uia___vals["Text"] = 50020
uia___fullnames["ToolBarControlTypeId"] = 50021
uia___vals["ToolBar"] = 50021
uia___fullnames["ToolTipControlTypeId"] = 50022
uia___vals["ToolTip"] = 50022
uia___fullnames["TreeControlTypeId"] = 50023
uia___vals["Tree"] = 50023
uia___fullnames["TreeItemControlTypeId"] = 50024
uia___vals["TreeItem"] = 50024
uia___fullnames["CustomControlTypeId"] = 50025
uia___vals["Custom"] = 50025
uia___fullnames["GroupControlTypeId"] = 50026
uia___vals["Group"] = 50026
uia___fullnames["ThumbControlTypeId"] = 50027
uia___vals["Thumb"] = 50027
uia___fullnames["DataGridControlTypeId"] = 50028
uia___vals["DataGrid"] = 50028
uia___fullnames["DataItemControlTypeId"] = 50029
uia___vals["DataItem"] = 50029
uia___fullnames["DocumentControlTypeId"] = 50030
uia___vals["Document"] = 50030
uia___fullnames["SplitButtonControlTypeId"] = 50031
uia___vals["SplitButton"] = 50031
uia___fullnames["WindowControlTypeId"] = 50032
uia___vals["Window"] = 50032
uia___fullnames["PaneControlTypeId"] = 50033
uia___vals["Pane"] = 50033
uia___fullnames["HeaderControlTypeId"] = 50034
uia___vals["Header"] = 50034
uia___fullnames["HeaderItemControlTypeId"] = 50035
uia___vals["HeaderItem"] = 50035
uia___fullnames["TableControlTypeId"] = 50036
uia___vals["Table"] = 50036
uia___fullnames["TitleBarControlTypeId"] = 50037
uia___vals["TitleBar"] = 50037
uia___fullnames["SeparatorControlTypeId"] = 50038
uia___vals["Separator"] = 50038
uia___fullnames["SemanticZoomControlTypeId"] = 50039
uia___vals["SemanticZoom"] = 50039
uia___fullnames["AppBarControlTypeId"] = 50040
uia___vals["AppBar"] = 50040
uia___fullnames["RuntimeIdPropertyId"] = 30000
uia___props["RuntimeId"] = 30000
uia___fullnames["BoundingRectanglePropertyId"] = 30001
uia___props["BoundingRectangle"] = 30001
uia___fullnames["ProcessIdPropertyId"] = 30002
uia___props["ProcessId"] = 30002
uia___fullnames["ControlTypePropertyId"] = 30003
uia___props["ControlType"] = 30003
uia___fullnames["LocalizedControlTypePropertyId"] = 30004
uia___props["LocalizedControlType"] = 30004
uia___fullnames["NamePropertyId"] = 30005
uia___props["Name"] = 30005
uia___fullnames["AcceleratorKeyPropertyId"] = 30006
uia___props["AcceleratorKey"] = 30006
uia___fullnames["AccessKeyPropertyId"] = 30007
uia___props["AccessKey"] = 30007
uia___fullnames["HasKeyboardFocusPropertyId"] = 30008
uia___props["HasKeyboardFocus"] = 30008
uia___fullnames["IsKeyboardFocusablePropertyId"] = 30009
uia___props["IsKeyboardFocusable"] = 30009
uia___fullnames["IsEnabledPropertyId"] = 30010
uia___props["IsEnabled"] = 30010
uia___fullnames["AutomationIdPropertyId"] = 30011
uia___props["AutomationId"] = 30011
uia___fullnames["ClassNamePropertyId"] = 30012
uia___props["ClassName"] = 30012
uia___fullnames["HelpTextPropertyId"] = 30013
uia___props["HelpText"] = 30013
uia___fullnames["ClickablePointPropertyId"] = 30014
uia___props["ClickablePoint"] = 30014
uia___fullnames["CulturePropertyId"] = 30015
uia___props["Culture"] = 30015
uia___fullnames["IsControlElementPropertyId"] = 30016
uia___props["IsControlElement"] = 30016
uia___fullnames["IsContentElementPropertyId"] = 30017
uia___props["IsContentElement"] = 30017
uia___fullnames["LabeledByPropertyId"] = 30018
uia___props["LabeledBy"] = 30018
uia___fullnames["IsPasswordPropertyId"] = 30019
uia___props["IsPassword"] = 30019
uia___fullnames["NativeWindowHandlePropertyId"] = 30020
uia___props["NativeWindowHandle"] = 30020
uia___fullnames["ItemTypePropertyId"] = 30021
uia___props["ItemType"] = 30021
uia___fullnames["IsOffscreenPropertyId"] = 30022
uia___props["IsOffscreen"] = 30022
uia___fullnames["OrientationPropertyId"] = 30023
uia___props["Orientation"] = 30023
uia___fullnames["FrameworkIdPropertyId"] = 30024
uia___props["FrameworkId"] = 30024
uia___fullnames["IsRequiredForFormPropertyId"] = 30025
uia___props["IsRequiredForForm"] = 30025
uia___fullnames["ItemStatusPropertyId"] = 30026
uia___props["ItemStatus"] = 30026
uia___fullnames["IsDockPatternAvailablePropertyId"] = 30027
uia___props["IsDockPatternAvailable"] = 30027
uia___fullnames["IsExpandCollapsePatternAvailablePropertyId"] = 30028
uia___props["IsExpandCollapsePatternAvailable"] = 30028
uia___fullnames["IsGridItemPatternAvailablePropertyId"] = 30029
uia___props["IsGridItemPatternAvailable"] = 30029
uia___fullnames["IsGridPatternAvailablePropertyId"] = 30030
uia___props["IsGridPatternAvailable"] = 30030
uia___fullnames["IsInvokePatternAvailablePropertyId"] = 30031
uia___props["IsInvokePatternAvailable"] = 30031
uia___fullnames["IsMultipleViewPatternAvailablePropertyId"] = 30032
uia___props["IsMultipleViewPatternAvailable"] = 30032
uia___fullnames["IsRangeValuePatternAvailablePropertyId"] = 30033
uia___props["IsRangeValuePatternAvailable"] = 30033
uia___fullnames["IsScrollPatternAvailablePropertyId"] = 30034
uia___props["IsScrollPatternAvailable"] = 30034
uia___fullnames["IsScrollItemPatternAvailablePropertyId"] = 30035
uia___props["IsScrollItemPatternAvailable"] = 30035
uia___fullnames["IsSelectionItemPatternAvailablePropertyId"] = 30036
uia___props["IsSelectionItemPatternAvailable"] = 30036
uia___fullnames["IsSelectionPatternAvailablePropertyId"] = 30037
uia___props["IsSelectionPatternAvailable"] = 30037
uia___fullnames["IsTablePatternAvailablePropertyId"] = 30038
uia___props["IsTablePatternAvailable"] = 30038
uia___fullnames["IsTableItemPatternAvailablePropertyId"] = 30039
uia___props["IsTableItemPatternAvailable"] = 30039
uia___fullnames["IsTextPatternAvailablePropertyId"] = 30040
uia___props["IsTextPatternAvailable"] = 30040
uia___fullnames["IsTogglePatternAvailablePropertyId"] = 30041
uia___props["IsTogglePatternAvailable"] = 30041
uia___fullnames["IsTransformPatternAvailablePropertyId"] = 30042
uia___props["IsTransformPatternAvailable"] = 30042
uia___fullnames["IsValuePatternAvailablePropertyId"] = 30043
uia___props["IsValuePatternAvailable"] = 30043
uia___fullnames["IsWindowPatternAvailablePropertyId"] = 30044
uia___props["IsWindowPatternAvailable"] = 30044
uia___fullnames["ValueValuePropertyId"] = 30045
uia___props["ValueValue"] = 30045
uia___fullnames["ValueIsReadOnlyPropertyId"] = 30046
uia___props["ValueIsReadOnly"] = 30046
uia___fullnames["RangeValueValuePropertyId"] = 30047
uia___props["RangeValueValue"] = 30047
uia___fullnames["RangeValueIsReadOnlyPropertyId"] = 30048
uia___props["RangeValueIsReadOnly"] = 30048
uia___fullnames["RangeValueMinimumPropertyId"] = 30049
uia___props["RangeValueMinimum"] = 30049
uia___fullnames["RangeValueMaximumPropertyId"] = 30050
uia___props["RangeValueMaximum"] = 30050
uia___fullnames["RangeValueLargeChangePropertyId"] = 30051
uia___props["RangeValueLargeChange"] = 30051
uia___fullnames["RangeValueSmallChangePropertyId"] = 30052
uia___props["RangeValueSmallChange"] = 30052
uia___fullnames["ScrollHorizontalScrollPercentPropertyId"] = 30053
uia___props["ScrollHorizontalScrollPercent"] = 30053
uia___fullnames["ScrollHorizontalViewSizePropertyId"] = 30054
uia___props["ScrollHorizontalViewSize"] = 30054
uia___fullnames["ScrollVerticalScrollPercentPropertyId"] = 30055
uia___props["ScrollVerticalScrollPercent"] = 30055
uia___fullnames["ScrollVerticalViewSizePropertyId"] = 30056
uia___props["ScrollVerticalViewSize"] = 30056
uia___fullnames["ScrollHorizontallyScrollablePropertyId"] = 30057
uia___props["ScrollHorizontallyScrollable"] = 30057
uia___fullnames["ScrollVerticallyScrollablePropertyId"] = 30058
uia___props["ScrollVerticallyScrollable"] = 30058
uia___fullnames["SelectionSelectionPropertyId"] = 30059
uia___props["SelectionSelection"] = 30059
uia___fullnames["SelectionCanSelectMultiplePropertyId"] = 30060
uia___props["SelectionCanSelectMultiple"] = 30060
uia___fullnames["SelectionIsSelectionRequiredPropertyId"] = 30061
uia___props["SelectionIsSelectionRequired"] = 30061
uia___fullnames["GridRowCountPropertyId"] = 30062
uia___props["GridRowCount"] = 30062
uia___fullnames["GridColumnCountPropertyId"] = 30063
uia___props["GridColumnCount"] = 30063
uia___fullnames["GridItemRowPropertyId"] = 30064
uia___props["GridItemRow"] = 30064
uia___fullnames["GridItemColumnPropertyId"] = 30065
uia___props["GridItemColumn"] = 30065
uia___fullnames["GridItemRowSpanPropertyId"] = 30066
uia___props["GridItemRowSpan"] = 30066
uia___fullnames["GridItemColumnSpanPropertyId"] = 30067
uia___props["GridItemColumnSpan"] = 30067
uia___fullnames["GridItemContainingGridPropertyId"] = 30068
uia___props["GridItemContainingGrid"] = 30068
uia___fullnames["DockDockPositionPropertyId"] = 30069
uia___props["DockDockPosition"] = 30069
uia___fullnames["ExpandCollapseExpandCollapseStatePropertyId"] = 30070
uia___props["ExpandCollapseExpandCollapseState"] = 30070
uia___fullnames["MultipleViewCurrentViewPropertyId"] = 30071
uia___props["MultipleViewCurrentView"] = 30071
uia___fullnames["MultipleViewSupportedViewsPropertyId"] = 30072
uia___props["MultipleViewSupportedViews"] = 30072
uia___fullnames["WindowCanMaximizePropertyId"] = 30073
uia___props["WindowCanMaximize"] = 30073
uia___fullnames["WindowCanMinimizePropertyId"] = 30074
uia___props["WindowCanMinimize"] = 30074
uia___fullnames["WindowWindowVisualStatePropertyId"] = 30075
uia___props["WindowWindowVisualState"] = 30075
uia___fullnames["WindowWindowInteractionStatePropertyId"] = 30076
uia___props["WindowWindowInteractionState"] = 30076
uia___fullnames["WindowIsModalPropertyId"] = 30077
uia___props["WindowIsModal"] = 30077
uia___fullnames["WindowIsTopmostPropertyId"] = 30078
uia___props["WindowIsTopmost"] = 30078
uia___fullnames["SelectionItemIsSelectedPropertyId"] = 30079
uia___props["SelectionItemIsSelected"] = 30079
uia___fullnames["SelectionItemSelectionContainerPropertyId"] = 30080
uia___props["SelectionItemSelectionContainer"] = 30080
uia___fullnames["TableRowHeadersPropertyId"] = 30081
uia___props["TableRowHeaders"] = 30081
uia___fullnames["TableColumnHeadersPropertyId"] = 30082
uia___props["TableColumnHeaders"] = 30082
uia___fullnames["TableRowOrColumnMajorPropertyId"] = 30083
uia___props["TableRowOrColumnMajor"] = 30083
uia___fullnames["TableItemRowHeaderItemsPropertyId"] = 30084
uia___props["TableItemRowHeaderItems"] = 30084
uia___fullnames["TableItemColumnHeaderItemsPropertyId"] = 30085
uia___props["TableItemColumnHeaderItems"] = 30085
uia___fullnames["ToggleToggleStatePropertyId"] = 30086
uia___props["ToggleToggleState"] = 30086
uia___fullnames["TransformCanMovePropertyId"] = 30087
uia___props["TransformCanMove"] = 30087
uia___fullnames["TransformCanResizePropertyId"] = 30088
uia___props["TransformCanResize"] = 30088
uia___fullnames["TransformCanRotatePropertyId"] = 30089
uia___props["TransformCanRotate"] = 30089
uia___fullnames["IsLegacyIAccessiblePatternAvailablePropertyId"] = 30090
uia___props["IsLegacyIAccessiblePatternAvailable"] = 30090
uia___fullnames["LegacyIAccessibleChildIdPropertyId"] = 30091
uia___props["LegacyIAccessibleChildId"] = 30091
uia___fullnames["LegacyIAccessibleNamePropertyId"] = 30092
uia___props["LegacyIAccessibleName"] = 30092
uia___fullnames["LegacyIAccessibleValuePropertyId"] = 30093
uia___props["LegacyIAccessibleValue"] = 30093
uia___fullnames["LegacyIAccessibleDescriptionPropertyId"] = 30094
uia___props["LegacyIAccessibleDescription"] = 30094
uia___fullnames["LegacyIAccessibleRolePropertyId"] = 30095
uia___props["LegacyIAccessibleRole"] = 30095
uia___fullnames["LegacyIAccessibleStatePropertyId"] = 30096
uia___props["LegacyIAccessibleState"] = 30096
uia___fullnames["LegacyIAccessibleHelpPropertyId"] = 30097
uia___props["LegacyIAccessibleHelp"] = 30097
uia___fullnames["LegacyIAccessibleKeyboardShortcutPropertyId"] = 30098
uia___props["LegacyIAccessibleKeyboardShortcut"] = 30098
uia___fullnames["LegacyIAccessibleSelectionPropertyId"] = 30099
uia___props["LegacyIAccessibleSelection"] = 30099
uia___fullnames["LegacyIAccessibleDefaultActionPropertyId"] = 30100
uia___props["LegacyIAccessibleDefaultAction"] = 30100
uia___fullnames["AriaRolePropertyId"] = 30101
uia___props["AriaRole"] = 30101
uia___fullnames["AriaPropertiesPropertyId"] = 30102
uia___props["AriaProperties"] = 30102
uia___fullnames["IsDataValidForFormPropertyId"] = 30103
uia___props["IsDataValidForForm"] = 30103
uia___fullnames["ControllerForPropertyId"] = 30104
uia___props["ControllerFor"] = 30104
uia___fullnames["DescribedByPropertyId"] = 30105
uia___props["DescribedBy"] = 30105
uia___fullnames["FlowsToPropertyId"] = 30106
uia___props["FlowsTo"] = 30106
uia___fullnames["ProviderDescriptionPropertyId"] = 30107
uia___props["ProviderDescription"] = 30107
uia___fullnames["IsItemContainerPatternAvailablePropertyId"] = 30108
uia___props["IsItemContainerPatternAvailable"] = 30108
uia___fullnames["IsVirtualizedItemPatternAvailablePropertyId"] = 30109
uia___props["IsVirtualizedItemPatternAvailable"] = 30109
uia___fullnames["IsSynchronizedInputPatternAvailablePropertyId"] = 30110
uia___props["IsSynchronizedInputPatternAvailable"] = 30110
uia___fullnames["OptimizeForVisualContentPropertyId"] = 30111
uia___props["OptimizeForVisualContent"] = 30111
uia___fullnames["IsObjectModelPatternAvailablePropertyId"] = 30112
uia___props["IsObjectModelPatternAvailable"] = 30112
uia___fullnames["AnnotationAnnotationTypeIdPropertyId"] = 30113
uia___props["AnnotationAnnotationTypeId"] = 30113
uia___fullnames["AnnotationAnnotationTypeNamePropertyId"] = 30114
uia___props["AnnotationAnnotationTypeName"] = 30114
uia___fullnames["AnnotationAuthorPropertyId"] = 30115
uia___props["AnnotationAuthor"] = 30115
uia___fullnames["AnnotationDateTimePropertyId"] = 30116
uia___props["AnnotationDateTime"] = 30116
uia___fullnames["AnnotationTargetPropertyId"] = 30117
uia___props["AnnotationTarget"] = 30117
uia___fullnames["IsAnnotationPatternAvailablePropertyId"] = 30118
uia___props["IsAnnotationPatternAvailable"] = 30118
uia___fullnames["IsTextPattern2AvailablePropertyId"] = 30119
uia___props["IsTextPattern2Available"] = 30119
uia___fullnames["StylesStyleIdPropertyId"] = 30120
uia___props["StylesStyleId"] = 30120
uia___fullnames["StylesStyleNamePropertyId"] = 30121
uia___props["StylesStyleName"] = 30121
uia___fullnames["StylesFillColorPropertyId"] = 30122
uia___props["StylesFillColor"] = 30122
uia___fullnames["StylesFillPatternStylePropertyId"] = 30123
uia___props["StylesFillPatternStyle"] = 30123
uia___fullnames["StylesShapePropertyId"] = 30124
uia___props["StylesShape"] = 30124
uia___fullnames["StylesFillPatternColorPropertyId"] = 30125
uia___props["StylesFillPatternColor"] = 30125
uia___fullnames["StylesExtendedPropertiesPropertyId"] = 30126
uia___props["StylesExtendedProperties"] = 30126
uia___fullnames["IsStylesPatternAvailablePropertyId"] = 30127
uia___props["IsStylesPatternAvailable"] = 30127
uia___fullnames["IsSpreadsheetPatternAvailablePropertyId"] = 30128
uia___props["IsSpreadsheetPatternAvailable"] = 30128
uia___fullnames["SpreadsheetItemFormulaPropertyId"] = 30129
uia___props["SpreadsheetItemFormula"] = 30129
uia___fullnames["SpreadsheetItemAnnotationObjectsPropertyId"] = 30130
uia___props["SpreadsheetItemAnnotationObjects"] = 30130
uia___fullnames["SpreadsheetItemAnnotationTypesPropertyId"] = 30131
uia___props["SpreadsheetItemAnnotationTypes"] = 30131
uia___fullnames["IsSpreadsheetItemPatternAvailablePropertyId"] = 30132
uia___props["IsSpreadsheetItemPatternAvailable"] = 30132
uia___fullnames["Transform2CanZoomPropertyId"] = 30133
uia___props["Transform2CanZoom"] = 30133
uia___fullnames["IsTransformPattern2AvailablePropertyId"] = 30134
uia___props["IsTransformPattern2Available"] = 30134
uia___fullnames["LiveSettingPropertyId"] = 30135
uia___props["LiveSetting"] = 30135
uia___fullnames["IsTextChildPatternAvailablePropertyId"] = 30136
uia___props["IsTextChildPatternAvailable"] = 30136
uia___fullnames["IsDragPatternAvailablePropertyId"] = 30137
uia___props["IsDragPatternAvailable"] = 30137
uia___fullnames["DragIsGrabbedPropertyId"] = 30138
uia___props["DragIsGrabbed"] = 30138
uia___fullnames["DragDropEffectPropertyId"] = 30139
uia___props["DragDropEffect"] = 30139
uia___fullnames["DragDropEffectsPropertyId"] = 30140
uia___props["DragDropEffects"] = 30140
uia___fullnames["IsDropTargetPatternAvailablePropertyId"] = 30141
uia___props["IsDropTargetPatternAvailable"] = 30141
uia___fullnames["DropTargetDropTargetEffectPropertyId"] = 30142
uia___props["DropTargetDropTargetEffect"] = 30142
uia___fullnames["DropTargetDropTargetEffectsPropertyId"] = 30143
uia___props["DropTargetDropTargetEffects"] = 30143
uia___fullnames["DragGrabbedItemsPropertyId"] = 30144
uia___props["DragGrabbedItems"] = 30144
uia___fullnames["Transform2ZoomLevelPropertyId"] = 30145
uia___props["Transform2ZoomLevel"] = 30145
uia___fullnames["Transform2ZoomMinimumPropertyId"] = 30146
uia___props["Transform2ZoomMinimum"] = 30146
uia___fullnames["Transform2ZoomMaximumPropertyId"] = 30147
uia___props["Transform2ZoomMaximum"] = 30147
uia___fullnames["FlowsFromPropertyId"] = 30148
uia___props["FlowsFrom"] = 30148
uia___fullnames["IsTextEditPatternAvailablePropertyId"] = 30149
uia___props["IsTextEditPatternAvailable"] = 30149
uia___fullnames["IsPeripheralPropertyId"] = 30150
uia___props["IsPeripheral"] = 30150
uia___fullnames["PositionInSetPropertyId"] = 30152
uia___props["PositionInSet"] = 30152
uia___fullnames["SizeOfSetPropertyId"] = 30153
uia___props["SizeOfSet"] = 30153
uia___fullnames["LevelPropertyId"] = 30154
uia___props["Level"] = 30154
uia___fullnames["AnnotationTypesPropertyId"] = 30155
uia___props["AnnotationTypes"] = 30155
uia___fullnames["AnnotationObjectsPropertyId"] = 30156
uia___props["AnnotationObjects"] = 30156
uia___fullnames["LandmarkTypePropertyId"] = 30157
uia___props["LandmarkType"] = 30157
uia___fullnames["LocalizedLandmarkTypePropertyId"] = 30158
uia___props["LocalizedLandmarkType"] = 30158
uia___fullnames["FullDescriptionPropertyId"] = 30159
uia___props["FullDescription"] = 30159
uia___fullnames["IsCustomNavigationPatternAvailablePropertyId"] = 30151
uia___props["IsCustomNavigationPatternAvailable"] = 30151
uia___fullnames["FillColorPropertyId"] = 30160
uia___props["FillColor"] = 30160
uia___fullnames["OutlineColorPropertyId"] = 30161
uia___props["OutlineColor"] = 30161
uia___fullnames["FillTypePropertyId"] = 30162
uia___props["FillType"] = 30162
uia___fullnames["VisualEffectsPropertyId"] = 30163
uia___props["VisualEffects"] = 30163
uia___fullnames["OutlineThicknessPropertyId"] = 30164
uia___props["OutlineThickness"] = 30164
uia___fullnames["CenterPointPropertyId"] = 30165
uia___props["CenterPoint"] = 30165
uia___fullnames["RotationPropertyId"] = 30166
uia___props["Rotation"] = 30166
uia___fullnames["SizePropertyId"] = 30167
uia___props["Size"] = 30167
uia___fullnames["UIA_AnnotationType_Unknown"] = 60000
uia___vals["UIA_AnnotationType_Unknown"] = 60000
uia___fullnames["UIA_AnnotationType_SpellingError"] = 60001
uia___vals["UIA_AnnotationType_SpellingError"] = 60001
uia___fullnames["UIA_AnnotationType_GrammarError"] = 60002
uia___vals["UIA_AnnotationType_GrammarError"] = 60002
uia___fullnames["UIA_AnnotationType_Comment"] = 60003
uia___vals["UIA_AnnotationType_Comment"] = 60003
uia___fullnames["UIA_AnnotationType_FormulaError"] = 60004
uia___vals["UIA_AnnotationType_FormulaError"] = 60004
uia___fullnames["UIA_AnnotationType_TrackChanges"] = 60005
uia___vals["UIA_AnnotationType_TrackChanges"] = 60005
uia___fullnames["UIA_AnnotationType_Header"] = 60006
uia___vals["UIA_AnnotationType_Header"] = 60006
uia___fullnames["UIA_AnnotationType_Footer"] = 60007
uia___vals["UIA_AnnotationType_Footer"] = 60007
uia___fullnames["UIA_AnnotationType_Highlighted"] = 60008
uia___vals["UIA_AnnotationType_Highlighted"] = 60008
uia___fullnames["UIA_AnnotationType_Endnote"] = 60009
uia___vals["UIA_AnnotationType_Endnote"] = 60009
uia___fullnames["UIA_AnnotationType_Footnote"] = 60010
uia___vals["UIA_AnnotationType_Footnote"] = 60010
uia___fullnames["UIA_AnnotationType_InsertionChange"] = 60011
uia___vals["UIA_AnnotationType_InsertionChange"] = 60011
uia___fullnames["UIA_AnnotationType_DeletionChange"] = 60012
uia___vals["UIA_AnnotationType_DeletionChange"] = 60012
uia___fullnames["UIA_AnnotationType_MoveChange"] = 60013
uia___vals["UIA_AnnotationType_MoveChange"] = 60013
uia___fullnames["UIA_AnnotationType_FormatChange"] = 60014
uia___vals["UIA_AnnotationType_FormatChange"] = 60014
uia___fullnames["UIA_AnnotationType_UnsyncedChange"] = 60015
uia___vals["UIA_AnnotationType_UnsyncedChange"] = 60015
uia___fullnames["UIA_AnnotationType_EditingLockedChange"] = 60016
uia___vals["UIA_AnnotationType_EditingLockedChange"] = 60016
uia___fullnames["UIA_AnnotationType_ExternalChange"] = 60017
uia___vals["UIA_AnnotationType_ExternalChange"] = 60017
uia___fullnames["UIA_AnnotationType_ConflictingChange"] = 60018
uia___vals["UIA_AnnotationType_ConflictingChange"] = 60018
uia___fullnames["UIA_AnnotationType_Author"] = 60019
uia___vals["UIA_AnnotationType_Author"] = 60019
uia___fullnames["UIA_AnnotationType_AdvancedProofingIssue"] = 60020
uia___vals["UIA_AnnotationType_AdvancedProofingIssue"] = 60020
uia___fullnames["UIA_AnnotationType_DataValidationError"] = 60021
uia___vals["UIA_AnnotationType_DataValidationError"] = 60021
uia___fullnames["UIA_AnnotationType_CircularReferenceError"] = 60022
uia___vals["UIA_AnnotationType_CircularReferenceError"] = 60022
uia___fullnames["ToolTipOpenedEventId"] = 20000
uia___events["ToolTipOpened"] = 20000
uia___fullnames["ToolTipClosedEventId"] = 20001
uia___events["ToolTipClosed"] = 20001
uia___fullnames["StructureChangedEventId"] = 20002
uia___events["StructureChanged"] = 20002
uia___fullnames["MenuOpenedEventId"] = 20003
uia___events["MenuOpened"] = 20003
uia___fullnames["AutomationPropertyChangedEventId"] = 20004
uia___events["AutomationPropertyChanged"] = 20004
uia___fullnames["AutomationFocusChangedEventId"] = 20005
uia___events["AutomationFocusChanged"] = 20005
uia___fullnames["AsyncContentLoadedEventId"] = 20006
uia___events["AsyncContentLoaded"] = 20006
uia___fullnames["MenuClosedEventId"] = 20007
uia___events["MenuClosed"] = 20007
uia___fullnames["LayoutInvalidatedEventId"] = 20008
uia___events["LayoutInvalidated"] = 20008
uia___fullnames["Invoke_InvokedEventId"] = 20009
uia___events["Invoke_Invoked"] = 20009
uia___fullnames["SelectionItem_ElementAddedToSelectionEventId"] = 20010
uia___events["SelectionItem_ElementAddedToSelection"] = 20010
uia___fullnames["SelectionItem_ElementRemovedFromSelectionEventId"] = 20011
uia___events["SelectionItem_ElementRemovedFromSelection"] = 20011
uia___fullnames["SelectionItem_ElementSelectedEventId"] = 20012
uia___events["SelectionItem_ElementSelected"] = 20012
uia___fullnames["Selection_InvalidatedEventId"] = 20013
uia___events["Selection_Invalidated"] = 20013
uia___fullnames["Text_TextSelectionChangedEventId"] = 20014
uia___events["Text_TextSelectionChanged"] = 20014
uia___fullnames["Text_TextChangedEventId"] = 20015
uia___events["Text_TextChanged"] = 20015
uia___fullnames["Window_WindowOpenedEventId"] = 20016
uia___events["Window_WindowOpened"] = 20016
uia___fullnames["Window_WindowClosedEventId"] = 20017
uia___events["Window_WindowClosed"] = 20017
uia___fullnames["MenuModeStartEventId"] = 20018
uia___events["MenuModeStart"] = 20018
uia___fullnames["MenuModeEndEventId"] = 20019
uia___events["MenuModeEnd"] = 20019
uia___fullnames["InputReachedTargetEventId"] = 20020
uia___events["InputReachedTarget"] = 20020
uia___fullnames["InputReachedOtherElementEventId"] = 20021
uia___events["InputReachedOtherElement"] = 20021
uia___fullnames["InputDiscardedEventId"] = 20022
uia___events["InputDiscarded"] = 20022
uia___fullnames["SystemAlertEventId"] = 20023
uia___events["SystemAlert"] = 20023
uia___fullnames["LiveRegionChangedEventId"] = 20024
uia___events["LiveRegionChanged"] = 20024
uia___fullnames["HostedFragmentRootsInvalidatedEventId"] = 20025
uia___events["HostedFragmentRootsInvalidated"] = 20025
uia___fullnames["DragStartEventId"] = 20026
uia___events["DragStart"] = 20026
uia___fullnames["DragCancelEventId"] = 20027
uia___events["DragCancel"] = 20027
uia___fullnames["Drag_CompleteEventId"] = 20028
uia___events["Drag_Complete"] = 20028
uia___fullnames["DragEnterEventId"] = 20029
uia___events["DragEnter"] = 20029
uia___fullnames["DragLeaveEventId"] = 20030
uia___events["DragLeave"] = 20030
uia___fullnames["DroppedEventId"] = 20031
uia___events["Dropped"] = 20031
uia___fullnames["TextEdit_TextChangedEventId"] = 20032
uia___events["TextEdit_TextChanged"] = 20032
uia___fullnames["TextEdit_ConversionTargetChangedEventId"] = 20033
uia___events["TextEdit_ConversionTargetChanged"] = 20033
uia___fullnames["OrientationType_None"] = 0
uia___vals["OrientationType_None"] = 0
uia___fullnames["OrientationType_Horizontal"] = 1
uia___vals["OrientationType_Horizontal"] = 1
uia___fullnames["OrientationType_Vertical"] = 2
uia___vals["OrientationType_Vertical"] = 2
uia___fullnames["TreeScope_Element"] = 1
uia___vals["TreeScope_Element"] = 1
uia___fullnames["TreeScope_Children"] = 2
uia___vals["TreeScope_Children"] = 2
uia___fullnames["TreeScope_Descendants"] = 4
uia___vals["TreeScope_Descendants"] = 4
uia___fullnames["TreeScope_Parent"] = 8
uia___vals["TreeScope_Parent"] = 8
uia___fullnames["TreeScope_Ancestors"] = 16
uia___vals["TreeScope_Ancestors"] = 16
uia___fullnames["TreeScope_Subtree"] = 7
uia___vals["TreeScope_Subtree"] = 7
uia___fullnames["next"] = 1
uia___vals["next"] = 1
uia___fullnames["prior"] = 0
uia___vals["prior"] = 0
uia___fullnames["ScrollAmount_LargeDecrement"] = 0
uia___vals["ScrollAmount_LargeDecrement"] = 0
uia___fullnames["ScrollAmount_SmallDecrement"] = 1
uia___vals["ScrollAmount_SmallDecrement"] = 1
uia___fullnames["ScrollAmount_NoAmount"] = 2
uia___vals["ScrollAmount_NoAmount"] = 2
uia___fullnames["ScrollAmount_LargeIncrement"] = 3
uia___vals["ScrollAmount_LargeIncrement"] = 3
uia___fullnames["ScrollAmount_SmallIncrement"] = 4
uia___vals["ScrollAmount_SmallIncrement"] = 4
uia___fullnames["StructureChangeType_ChildAdded"] = 0
uia___vals["StructureChangeType_ChildAdded"] = 0
uia___fullnames["StructureChangeType_ChildRemoved"] = 1
uia___vals["StructureChangeType_ChildRemoved"] = 1
uia___fullnames["StructureChangeType_ChildrenInvalidated"] = 2
uia___vals["StructureChangeType_ChildrenInvalidated"] = 2
uia___fullnames["StructureChangeType_ChildrenBulkAdded"] = 3
uia___vals["StructureChangeType_ChildrenBulkAdded"] = 3
uia___fullnames["StructureChangeType_ChildrenBulkRemoved"] = 4
uia___vals["StructureChangeType_ChildrenBulkRemoved"] = 4
uia___fullnames["StructureChangeType_ChildrenReordered"] = 5
uia___vals["StructureChangeType_ChildrenReordered"] = 5
uia___fullnames["AnimationStyleAttributeId"] = 40000
uia___attributes["AnimationStyle"] = 40000
uia___fullnames["AnnotationObjectsAttributeId"] = 40032
uia___attributes["AnnotationObjects"] = 40032
uia___fullnames["AnnotationTypesAttributeId"] = 40031
uia___attributes["AnnotationTypes"] = 40031
uia___fullnames["BackgroundColorAttributeId"] = 40001
uia___attributes["BackgroundColor"] = 40001
uia___fullnames["BulletStyleAttributeId"] = 40002
uia___attributes["BulletStyle"] = 40002
uia___fullnames["CapStyleAttributeId"] = 40003
uia___attributes["CapStyle"] = 40003
uia___fullnames["CaretBidiModeAttributeId"] = 40039
uia___attributes["CaretBidiMode"] = 40039
uia___fullnames["CaretPositionAttributeId"] = 40038
uia___attributes["CaretPosition"] = 40038
uia___fullnames["CultureAttributeId"] = 40004
uia___attributes["Culture"] = 40004
uia___fullnames["FontNameAttributeId"] = 40005
uia___attributes["FontName"] = 40005
uia___fullnames["FontSizeAttributeId"] = 40006
uia___attributes["FontSize"] = 40006
uia___fullnames["FontWeightAttributeId"] = 400007
uia___attributes["FontWeight"] = 400007
uia___fullnames["ForegroundColorAttributeId"] = 40008
uia___attributes["ForegroundColor"] = 40008
uia___fullnames["HorizontalTextAlignmentAttributeId"] = 40009
uia___attributes["HorizontalTextAlignment"] = 40009
uia___fullnames["IndentationFirstLineAttributeId"] = 40010
uia___attributes["IndentationFirstLine"] = 40010
uia___fullnames["IndentationLeadingAttributeId"] = 40011
uia___attributes["IndentationLeading"] = 40011
uia___fullnames["IndentationTrailingAttributeId"] = 40012
uia___attributes["IndentationTrailing"] = 40012
uia___fullnames["IsActiveAttributeId"] = 40036
uia___attributes["IsActive"] = 40036
uia___fullnames["IsHiddenAttributeId"] = 40013
uia___attributes["IsHidden"] = 40013
uia___fullnames["IsItalicAttributeId"] = 40014
uia___attributes["IsItalic"] = 40014
uia___fullnames["IsReadOnlyAttributeId"] = 40015
uia___attributes["IsReadOnly"] = 40015
uia___fullnames["IsSubscriptAttributeId"] = 40016
uia___attributes["IsSubscript"] = 40016
uia___fullnames["IsSuperscriptAttributeId"] = 40017
uia___attributes["IsSuperscript"] = 40017
uia___fullnames["LinkAttributeId"] = 40035
uia___attributes["Link"] = 40035
uia___fullnames["MarginBottomAttributeId"] = 40018
uia___attributes["MarginBottom"] = 40018
uia___fullnames["MarginLeadingAttributeId"] = 40019
uia___attributes["MarginLeading"] = 40019
uia___fullnames["MarginTopAttributeId"] = 40020
uia___attributes["MarginTop"] = 40020
uia___fullnames["MarginTrailingAttributeId"] = 40021
uia___attributes["MarginTrailing"] = 40021
uia___fullnames["OutlineStylesAttributeId"] = 40022
uia___attributes["OutlineStyles"] = 40022
uia___fullnames["OverlineColorAttributeId"] = 40023
uia___attributes["OverlineColor"] = 40023
uia___fullnames["OverlineStyleAttributeId"] = 40024
uia___attributes["OverlineStyle"] = 40024
uia___fullnames["SelectionActiveEndAttributeId"] = 40037
uia___attributes["SelectionActiveEnd"] = 40037
uia___fullnames["StrikethroughColorAttributeId"] = 40025
uia___attributes["StrikethroughColor"] = 40025
uia___fullnames["StrikethroughStyleAttributeId"] = 40026
uia___attributes["StrikethroughStyle"] = 40026
uia___fullnames["StyleIdAttributeId"] = 40034
uia___attributes["StyleId"] = 40034
uia___fullnames["StyleNameAttributeId"] = 40033
uia___attributes["StyleName"] = 40033
uia___fullnames["TabsAttributeId"] = 40027
uia___attributes["Tabs"] = 40027
uia___fullnames["TextFlowDirectionsAttributeId"] = 40028
uia___attributes["TextFlowDirections"] = 40028
uia___fullnames["UnderlineColorAttributeId"] = 40029
uia___attributes["UnderlineColor"] = 40029
uia___fullnames["UnderlineStyleAttributeId"] = 40030
uia___attributes["UnderlineStyle"] = 40030
uia___fullnames["TextUnit_Character"] = 0
uia___textUnits["Character"] = 0
uia___fullnames["TextUnit_Format"] = 1
uia___textUnits["Format"] = 1
uia___fullnames["TextUnit_Word"] = 2
uia___textUnits["Word"] = 2
uia___fullnames["TextUnit_Line"] = 3
uia___textUnits["Line"] = 3
uia___fullnames["TextUnit_Paragraph"] = 4
uia___textUnits["Paragraph"] = 4
uia___fullnames["TextUnit_Page"] = 5
uia___textUnits["Page"] = 5
uia___fullnames["TextUnit_Document"] = 6
uia___textUnits["Document"] = 6
uia___fullnames["TextPatternRangeEndpoint_Start"] = 0
uia___vals["TextPatternRangeEndpoint_Start"] = 0
uia___fullnames["TextPatternRangeEndpoint_End"] = 1
uia___vals["TextPatternRangeEndpoint_End"] = 1
uia___fullnames["SupportedTextSelection_None"] = 0
uia___vals["SupportedTextSelection_None"] = 0
uia___fullnames["SupportedTextSelection_Single"] = 1
uia___vals["SupportedTextSelection_Single"] = 1
uia___fullnames["SupportedTextSelection_Multiple"] = 2
uia___vals["SupportedTextSelection_Multiple"] = 2
uia___fullnames["DockPosition_Top"] = 0
uia___vals["DockPosition_Top"] = 0
uia___fullnames["DockPosition_Left"] = 1
uia___vals["DockPosition_Left"] = 1
uia___fullnames["DockPosition_Bottom"] = 2
uia___vals["DockPosition_Bottom"] = 2
uia___fullnames["DockPosition_Right"] = 3
uia___vals["DockPosition_Right"] = 3
uia___fullnames["DockPosition_Fill"] = 4
uia___vals["DockPosition_Fill"] = 4
uia___fullnames["DockPosition_None"] = 5
uia___vals["DockPosition_None"] = 5
uia___fullnames["SynchronizedInputType_KeyUp"] = 0
uia___vals["SynchronizedInputType_KeyUp"] = 0
uia___fullnames["SynchronizedInputType_KeyDown"] = 1
uia___vals["SynchronizedInputType_KeyDown"] = 1
uia___fullnames["SynchronizedInputType_LeftMouseUp"] = 2
uia___vals["SynchronizedInputType_LeftMouseUp"] = 2
uia___fullnames["SynchronizedInputType_LeftMouseDown"] = 3
uia___vals["SynchronizedInputType_LeftMouseDown"] = 3
uia___fullnames["SynchronizedInputType_RightMouseUp"] = 4
uia___vals["SynchronizedInputType_RightMouseUp"] = 4
uia___fullnames["SynchronizedInputType_RightMouseDown"] = 5
uia___vals["SynchronizedInputType_RightMouseDown"] = 5
uia___fullnames["RowOrColumnMajor_RowMajor"] = 0
uia___vals["RowOrColumnMajor_RowMajor"] = 0
uia___fullnames["RowOrColumnMajor_ColumnMajor"] = 1
uia___vals["RowOrColumnMajor_ColumnMajor"] = 1
uia___fullnames["RowOrColumnMajor_Indeterminate"] = 2
uia___vals["RowOrColumnMajor_Indeterminate"] = 2
uia___fullnames["ZoomUnit_NoAmount"] = 0
uia___vals["ZoomUnit_NoAmount"] = 0
uia___fullnames["ZoomUnit_LargeDecrement"] = 1
uia___vals["ZoomUnit_LargeDecrement"] = 1
uia___fullnames["ZoomUnit_SmallDecrement"] = 2
uia___vals["ZoomUnit_SmallDecrement"] = 2
uia___fullnames["ZoomUnit_LargeIncrement"] = 3
uia___vals["ZoomUnit_LargeIncrement"] = 3
uia___fullnames["ZoomUnit_SmallIncrement"] = 4
uia___vals["ZoomUnit_SmallIncrement"] = 4
uia___fullnames["WindowVisualState_Normal"] = 0
uia___vals["WindowVisualState_Normal"] = 0
uia___fullnames["WindowVisualState_Maximized"] = 1
uia___vals["WindowVisualState_Maximized"] = 1
uia___fullnames["WindowVisualState_Minimized"] = 2
uia___vals["WindowVisualState_Minimized"] = 2
uia___fullnames["WindowInteractionState_Running"] = 0
uia___vals["WindowInteractionState_Running"] = 0
uia___fullnames["WindowInteractionState_Closing"] = 1
uia___vals["WindowInteractionState_Closing"] = 1
uia___fullnames["WindowInteractionState_ReadyForUserInteraction"] = 2
uia___vals["WindowInteractionState_ReadyForUserInteraction"] = 2
uia___fullnames["WindowInteractionState_BlockedByModalWindow"] = 3
uia___vals["WindowInteractionState_BlockedByModalWindow"] = 3
uia___fullnames["WindowInteractionState_NotResponding"] = 4
uia___vals["WindowInteractionState_NotResponding"] = 4
uia___fullnames["TextEditChangeType_None"] = 0
uia___vals["TextEditChangeType_None"] = 0
uia___fullnames["TextEditChangeType_AutoCorrect"] = 1
uia___vals["TextEditChangeType_AutoCorrect"] = 1
uia___fullnames["TextEditChangeType_Composition"] = 2
uia___vals["TextEditChangeType_Composition"] = 2
uia___fullnames["TextEditChangeType_CompositionFinalized"] = 3
uia___vals["TextEditChangeType_CompositionFinalized"] = 3
uia___fullnames["ExpandCollapseState_Collapsed"] = 0
uia___vals["ExpandCollapseState_Collapsed"] = 0
uia___fullnames["ExpandCollapseState_Expanded"] = 1
uia___vals["ExpandCollapseState_Expanded"] = 1
uia___fullnames["ExpandCollapseState_PartiallyExpanded"] = 2
uia___vals["ExpandCollapseState_PartiallyExpanded"] = 2
uia___fullnames["ExpandCollapseState_LeafNode"] = 3
uia___vals["ExpandCollapseState_LeafNode"] = 3
uia___fullnames["StyleId_BulletedList"] = 70015
uia___vals["StyleId_BulletedList"] = 70015
uia___fullnames["StyleId_Custom"] = 70000
uia___vals["StyleId_Custom"] = 70000
uia___fullnames["StyleId_Emphasis"] = 70013
uia___vals["StyleId_Emphasis"] = 70013
uia___fullnames["StyleId_Heading1"] = 70001
uia___vals["StyleId_Heading1"] = 70001
uia___fullnames["StyleId_Heading2"] = 70002
uia___vals["StyleId_Heading2"] = 70002
uia___fullnames["StyleId_Heading3"] = 70003
uia___vals["StyleId_Heading3"] = 70003
uia___fullnames["StyleId_Heading4"] = 70004
uia___vals["StyleId_Heading4"] = 70004
uia___fullnames["StyleId_Heading5"] = 70005
uia___vals["StyleId_Heading5"] = 70005
uia___fullnames["StyleId_Heading6"] = 70006
uia___vals["StyleId_Heading6"] = 70006
uia___fullnames["StyleId_Heading7"] = 70007
uia___vals["StyleId_Heading7"] = 70007
uia___fullnames["StyleId_Heading8"] = 70008
uia___vals["StyleId_Heading8"] = 70008
uia___fullnames["StyleId_Heading9"] = 70009
uia___vals["StyleId_Heading9"] = 70009
uia___fullnames["StyleId_Normal"] = 70012
uia___vals["StyleId_Normal"] = 70012
uia___fullnames["StyleId_NumberedList"] = 70016
uia___vals["StyleId_NumberedList"] = 70016
uia___fullnames["StyleId_Quote"] = 70014
uia___vals["StyleId_Quote"] = 70014
uia___fullnames["StyleId_Subtitle"] = 70011
uia___vals["StyleId_Subtitle"] = 70011
uia___fullnames["StyleId_Title"] = 70010
uia___vals["StyleId_Title"] = 70010
uia___fullnames["UIA_CustomLandmarkTypeId"] = 80000
uia___vals["UIA_CustomLandmarkTypeId"] = 80000
uia___fullnames["UIA_FormLandmarkTypeId"] = 80001
uia___vals["UIA_FormLandmarkTypeId"] = 80001
uia___fullnames["UIA_MainLandmarkTypeId"] = 80002
uia___vals["UIA_MainLandmarkTypeId"] = 80002
uia___fullnames["UIA_NavigationLandmarkTypeId"] = 80003
uia___vals["UIA_NavigationLandmarkTypeId"] = 80003
uia___fullnames["UIA_SearchLandmarkTypeId"] = 80004
uia___vals["UIA_SearchLandmarkTypeId"] = 80004
uia___fullnames["ToggleState_Off"] = 0
uia___vals["ToggleState_Off"] = 0
uia___fullnames["ToggleState_On"] = 1
uia___vals["ToggleState_On"] = 1
uia___fullnames["ToggleState_Indeterminate"] = 2
uia___vals["ToggleState_Indeterminate"] = 2
uia___fullnames["Off"] = 0
uia___vals["Off"] = 0
uia___fullnames["Polite"] = 1
uia___vals["Polite"] = 1
uia___fullnames["Assertive"] = 2
uia___vals["Assertive"] = 2
uia___fullnames["SELFLAG_NONE"] = 0
uia___vals["SELFLAG_NONE"] = 0
uia___fullnames["SELFLAG_TAKEFOCUS"] = 1
uia___vals["SELFLAG_TAKEFOCUS"] = 1
uia___fullnames["SELFLAG_TAKESELECTION"] = 2
uia___vals["SELFLAG_TAKESELECTION"] = 2
uia___fullnames["SELFLAG_EXTENDSELECTION"] = 4
uia___vals["SELFLAG_EXTENDSELECTION"] = 4
uia___fullnames["SELFLAG_ADDSELECTION"] = 8
uia___vals["SELFLAG_ADDSELECTION"] = 8
uia___fullnames["SELFLAG_REMOVESELECTION"] = 16
uia___vals["SELFLAG_REMOVESELECTION"] = 16
uia___fullnames["TextRange_NoMaxLength"] = 0xffffffff
uia___vals["TextRange_NoMaxLength"] = 0xffffffff
uia___fullnames["CaretPosition_Unknown"] = 0
uia___vals["CaretPosition_Unknown"] = 0
uia___fullnames["CaretPosition_EndOfLine"] = 1
uia___vals["CaretPosition_EndOfLine"] = 1
uia___fullnames["CaretPosition_BeginningOfLine"] = 2
uia___vals["CaretPosition_BeginningOfLine"] = 2
uia___fullnames["ActiveEnd_None"] = 0
uia___vals["ActiveEnd_None"] = 0
uia___fullnames["ActiveEnd_Start"] = 1
uia___vals["ActiveEnd_Start"] = 1
uia___fullnames["ActiveEnd_End"] = 2
uia___vals["ActiveEnd_End"] = 2

uia___built = True
endFunction

collection function uia__propList()
; Return the list of supported property names.
; This is a collection with names as keys and integer property Ids as values.
uia___buildLists()
return uia___props
endFunction

void function uia___sayError(string s, optional string spec)
; Speak an error message.
if spec
	s = s +": " +spec
endIf
sayMessage(OT_Error, s)
endFunction


;---------------------------------- Module stl ---------------------------------
; Standard data structures for JAWS.
;
; Author: Doug Lee of Level Access

const
; List types supported.
	uia__stl___LstStack = 1,
	uia__stl___LstQueue = 2

collection function uia__stl__stack(optional int maxSize)
; Return a stack for use via uia__stl__push/uia__stl__pop etc.
return uia__stl___new(uia__stl___LstStack, maxSize)
endFunction

collection function uia__stl__queue(optional int maxSize)
; Return a queue for use via uia__stl__push/uia__stl__pop etc.
return uia__stl___new(uia__stl___LstQueue, maxSize)
endFunction

collection function uia__stl___new(int ltype, optional int maxSize)
; Return a structure of the indicated type.
var collection lst lst = new collection
lst.type = ltype
if ltype == uia__stl___LstQueue
	; ToDo: maxSize ignored.
	lst.size = 0
	return lst
endIf

lst.maxSize = 0
lst.top = 0
if maxSize
	; More efficient.
	lst.maxSize = maxSize
	var variantArray ar ar = new variantArray[lst.maxSize]
	lst.content = ar
else
	; A bit less efficient but more flexible.
	var collection nc nc = new collection
	lst.content = nc
endIf
return lst
endFunction

int function uia__stl__push(collection struct, variant item)
; Push an item onto whatever structure is passed.
; Returns True on success and False on failure.
if struct.type == uia__stl___LstQueue
	var collection node node = new collection
	node.value = item
	var collection node0 = struct.tail
	if node0
		node0.next = node
		struct.tail = node
	else
		struct.head = node
		struct.tail = node
	endIf
	struct.size = struct.size +1
	return True
endIf

struct.top = struct.top +1
if struct.maxSize
	if struct.top > struct.maxSize
		struct.top = struct.top -1
		return False
	endIf
	var variantArray ar = struct.content
	ar[struct.top] = item
	;sayString("Push " +intToString(struct.top))
else
	var collection col = struct.content
	col[intToString(struct.top)] = item
	;sayString("Dynamic push " +intToString(struct.top))
endIf
return True
endFunction

collection function uia__stl__top(collection struct, optional int alsoPop)
; Return the top element of whatever structure is passed.
; Also pops the element off if alsoPop is True.
; The item is returned as the .value property of a simple collection.
; On error, null is returned rather than a collection.
var collection item item = new collection
if struct.type == uia__stl___LstQueue
	var collection node = struct.head
	if !node
		return
	elif alsoPop
		struct.head = node.next
		if !struct.head
			struct.tail = struct.head
		endIf
		struct.size = struct.size -1
	endIf
	return node
endIf

if struct.top < 1
	return
endIf
if struct.type == uia__stl___LstStack
	if struct.maxSize
		var variantArray ar = struct.content
		item.value = ar[struct.top]
	else
		var collection col = struct.content
		item.value = col[intToString(struct.top)]
	endIf
	if alsoPop
		;sayString("Pop " +intToString(struct.top))
		struct.top = struct.top -1
	endIf
	return item
endIf
return
endFunction

collection function uia__stl__pop(collection struct)
; Pop an item from whatever structure is passed.
; The item comes back as the .value property of a collection.
; On error, returns null instead of a collection.
return uia__stl__top(struct, True)
endFunction

int function uia__stl__size(collection struct)
; Return the size (item count) of whatever structure is passed.
if struct.type == uia__stl___LstQueue
	return struct.size
endIf
return struct.top
endFunction

;-------------------------------- End module stl -------------------------------

;================================ End module uia ===============================


;================================= Module util =================================
; Miscellaneous utility functions
;
; Author:  Doug Lee

int function util__isSpecialFocus()
; Returns True if there is a special condition for which normal focus handling should not apply.
; Currently, the conditions that are considered special are menusActive, userBufferIsActive, inHJDialog, and switchingTasks (Alt+Tab is active).
return menusActive() ||userBufferIsActive() ||inHJDialog() || util__switchingTasks()
endFunction

int function util__switchingTasks()
; Returns True if the user is currently Alt+Tabbing between tasks.

; This approach is adapted from JAWS 9.0.2169's default.jss FocusChangedEventEx function.
if getWindowClass(getFocus()) == "#32771"
&& getObjectTypeCode()== WT_LISTBOXITEM
	return True
endIf
return False
endFunction

string function util__stringFromFile(string fname, /*?*/ string fmt)
; Returns the file contents as a string.
; fmt: null for system default, "a" for ASCII, "u" for Unicode.
var
 object fso, object f,
	int iFmt,
	string s
iFmt = -2  ; TristateUseDefault
if fmt == "a"
	iFmt = 0  ; TristateFalse
elif fmt == "u"
	iFmt = -1  ; TristateTrue
endIf
fso = createObjectEx("Scripting.FilesystemObject", False)
if !fso
fso = createObjectEx("Scripting.FilesystemObject", True)
endIf
if !fso
	;sayMessage(OT_Error, "Unable to get a fileSystemObject")
	return ""
endIf
f = fso.openTextFile(fname, 1, 0, iFmt)  ; 1 is ForReading, 0 means don't create
if !f
	;sayString("No code file")
	return ""
endIf
s = f.readAll()
f.close()
if !stringLength(s)
	;sayString("No code in the file")
endIf
return s
endFunction

int function util__stringToFile(string fname, string s, /*?*/ string fmt)
; Writes the string to the indicated file, creating or replacing the file.
; fmt: null for system default, "a" for ASCII, "u" for Unicode.
var
 object fso, object f,
	int iFmt
iFmt = -2  ; TristateUseDefault
if fmt == "a"
	iFmt = 0  ; TristateFalse
elif fmt == "u"
	iFmt = -1  ; TristateTrue
endIf
fso = createObjectEx("Scripting.FilesystemObject", False)
if !fso
fso = createObjectEx("Scripting.FilesystemObject", True)
endIf
if !fso
	;sayMessage(OT_Error, "Unable to get a fileSystemObject")
	return ""
endIf
f = fso.openTextFile(fname, 2, 1, iFmt)  ; 2 is ForWriting, 1 means create
if !f
	;sayString("Can't create file")
	return ""
endIf
f.write(s)
f.close()
return True
endFunction

collection function util__getWindowsVersionInfo()
; Return a structure of information about the running Windows version. Fields:
;	structVer: Integer indicating version of this structure, oldest being 1.
;	versionString: The version string of C:\Windows\Explorer.exe, uninterpreted.
;	major, minor, build, revision: The four standard parts of the version code as integers.
;	bits: 32, 64, etc., version indicator (integer).
;	osCode: A short string indicating Windows version and bit size, separated by an underscore. Example: Win8.1_64.
var string sVer = util__getFileVersion("C:\\Windows\\Explorer.exe")
if !sVer
	return
endIf
var collection info info = new collection
info.structVer = 1
info.versionString = sVer
info.major = stringToInt(stringSegment(sVer, ".", 1))
info.minor = stringToInt(stringSegment(sVer, ".", 2))
info.build = stringToInt(stringSegment(sVer, ".", 3))
info.revision = stringToInt(stringSegment(sVer, ".", 4))

; The following lookup table is derived from a Windows version table in the Virtual Audio Cable Glossary.
; [DGL, 2016-09-18]
var string sLookup = formatString("%1.%2", intToString(info.major), intToString(info.minor))
var string vers = "5.0|5.1|5.2|6.0|6.1|6.2|6.3|10.0"
var string codes = "Win2k|WinXP|Win2k3|WinVista|Win7|Win8|Win8.1|Win10"
var int idx = stringSegmentIndex(vers, "|", sLookup, True)
if idx
	info.osCode = stringSegment(codes, "|", idx)
elif info.major == 4
	; These could be further broken down but are not likely to be encountered in JAWS by now.
	info.osCode = "Win9x/ME"
else
	info.osCode = formatString("unknown(%1)", sLookup)
endIf
info.bits = 32
if fileExists("C:\\Windows\\SysWow64")
	info.bits = 64
endIf
info.osCode = formatString("%1_%2", info.osCode, intToString(info.bits))
return info
endFunction

string function util__getFileVersion(/*?*/ string appPath)
; Return the version number of the current or indicated program as a string.
; The string should be separated into fields by a dot (".").
; The first segment is the major number, then minor, then more minor etc.
; The specific number of segments varies by program.
if !appPath
	appPath = getAppFilePath()
endIf
var string sver
var object fso = util__getFSObject()
if fso
	sver = fso.getFileVersion(appPath)
endIf
if !sver
	sver = getVersionInfoString(appPath, "ProductVersion")
endIf
if !sver
	sver = getVersionInfoString(appPath, "FileVersion")
endIf
sver = stringReplaceSubstrings(sver, " ", "")
sver = stringReplaceSubstrings(sver, ",", ".")
return sver
endFunction

globals object util___fso
object function util__getFSObject()
; Cached access to a filesystem object.
if !util___fso
	util___fso = util__createObject("Scripting.FilesystemObject")
endIf
return util___fso
endFunction

object function util__createObject(string objname)
; A CreateObject wrapper that avoids causing 5-7 second delays in DOS boxes when called if JAWS is running as a service.
; As of June 11, 2011, uses the CoCreateInstance method of creating an object
; unless that doesn't work, whereupon the GetObjectHelper method is used.
; Previously the CoCreateInstance method was only used if JAWS was running as a service.
var object o = createObjectEx(objname, False)
if !o
	o = createObject(objname)
endIf
return o
endFunction

;=============================== End module util ===============================

